/**************************************************************************
 *                                                                        *
 *                       This file is part of PSReg                       *
 *                                                                        *
 *          Copyright (C) 2002 Ko Nishino. All Rights Reserved.           *
 *                                                                        *
 *  Permission to use this software for educational and research purposes *
 * is hereby granted without fee. All other rights, including the rights  *
 * to distribute, modify, or sell the software, are explicitly reserved   *
 * by the author. Please contact the author if you wish to make any       *
 * modifications or additions to the software.                            *
 *                                                                        *
 *  This software is provided "as is" and without warranty of any kind,   *
 * express, implied or otherwise, including without limitation, any       *
 * warranty of merchantability or fitness for a particular purpose.       *
 *                                                                        *
 **************************************************************************/
/*  
 *  3DFile.cpp
 *          Ko Nishino 
 *               kon@cvl.iis.u-tokyo.ac.jp
 *               http://www.cvl.iis.u-tokyo.ac.jp/~kon/
 */

#include "KDefs.h"
#include "3DFile.h"
#include "Utils.h"
#include <stdio.h>
#include <iostream>
#include <fstream>

#include <gsl/gsl_vector.h>

C3DFile::C3DFile()
{
    m_pVerts = NULL;
    m_pFaces = NULL;
    m_pcShortFName = m_pcFName = NULL;
    m_ulNVerts = m_ulNFaces = 0;
    m_pfMat = NULL;
    // many things missing
}

C3DFile::C3DFile(const C3DFile& c3dfile)
{
    std::cout << "copy constructer loaded" << std::endl;
    this->SetNVerts(c3dfile.GetNVerts());
    this->SetVerts(c3dfile.GetVerts());
    this->SetNFaces(c3dfile.GetNFaces());
    this->SetFaces(c3dfile.GetFaces());
// many things missing
    return;
}

C3DFile::~C3DFile()
{
    if(m_pVerts != NULL){
  	delete [] m_pVerts;
    }
    if(m_pFaces != NULL){
  	delete [] m_pFaces;
    }
    if(m_pcShortFName != NULL){
	delete [] m_pcShortFName;
    }
    if(m_pcFName != NULL){
	delete [] m_pcFName;
    }
    delete [] m_pfMat;

}

/* Be careful!!
   You always need to store the vertex ids in faces in an order as 
   for(x = 0; x < nWidth; x++){
       for(y = 0; y < nHeight; y++){

  */
void C3DFile::SetVerts(VRTX** pvrtx, int nWidth, int nHeight)
{
    if(m_ulNVerts == 0){
	std::cerr << "Set the number of vertices first." << std::endl;
	return;
    }
    assert(nWidth*nHeight == m_ulNVerts);
    if(m_pVerts == NULL){
	m_pVerts = new VRTX[m_ulNVerts];
    }
    int ncnt(0);
    // Keeping this order (raster scan) is SUPER IMPORTANT
    for(int y = 0; y < nHeight; y++){
	for(int x = 0; x < nWidth; x++){
	    m_pVerts[ncnt] = pvrtx[x][y];
	    ncnt++;
	}
    }
}

void C3DFile::Alloc()
{
    if(m_pVerts != NULL){
	delete [] m_pVerts;
    }
    if(m_ulNVerts == 0){
	std::cerr << "Set the nubmer of vertices first." << std::endl;
	return;
    }
    m_pVerts = new VRTX[m_ulNVerts];
    if(m_pFaces != NULL){
	delete [] m_pFaces;
    }
    if(m_ulNFaces == 0){
	std::cerr << "Set the number of faces first." << std::endl;
	return;
    }
    m_pFaces = new TRNGL[m_ulNFaces];
}

void C3DFile::SetVerts(VRTX* pvrtx)
{
    if(m_ulNVerts == 0){
	std::cerr << "Set the nubmer of vertices first." << std::endl;
	return;
    }
    if(m_pVerts == NULL){
	m_pVerts = new VRTX[m_ulNVerts];
    }
    for(unsigned long ul = 0; ul < m_ulNVerts; ul++){
	m_pVerts[ul] = pvrtx[ul];
    }
}

void C3DFile::SetFaces(TRNGL* ptrngls)
{
    if(m_ulNFaces == 0){
	std::cerr << "Set the number of faces first." << std::endl;
	return;
    }
    if(m_pFaces == NULL){
	m_pFaces = new TRNGL[m_ulNFaces];
    }

    for(unsigned long ul = 0; ul < m_ulNFaces; ul++){
	m_pFaces[ul] = ptrngls[ul];
    }

}

void C3DFile::SetFaces(std::vector<TRNGL>& vTrngls)
{
    if(m_ulNFaces == 0){
	std::cerr << "Set the number of faces first." << std::endl;
	return;
    }
    assert(vTrngls.size() == m_ulNFaces);
    if(m_pFaces == NULL){
	m_pFaces = new TRNGL[m_ulNFaces];
    }
    std::vector<TRNGL>::iterator itr;
    int ncnt(0);
    for(itr = vTrngls.begin(); itr < vTrngls.end(); itr++){
	m_pFaces[ncnt] = *itr;
	ncnt++;
    }

}

int C3DFile::WritePly(const char* pcFName, float fScale, bool bAscii, bool bLittle)
{
    FILE* fp;
    fp = fopen(pcFName, "w");
    if(NULL == fp){
	std::cerr << "Could not save PLY." << std::endl;
	return 1;
    }
    
    fprintf(fp, "ply\n");
    if(bAscii){
	fprintf(fp, "format ascii 1.0\n");
    }else if(bLittle){
	fprintf(fp, "format binary little endian\n");
    }else{
	fprintf(fp, "format binary big endian\n");
    }
    fprintf(fp, "comment %s\n", m_strCmnt.c_str());
    if(m_pfMat != NULL){
	for(int i = 0; i < 4; i++){
	    fprintf(fp, "matrix ");
	    for(int j = 0; j < 4; j++){
		fprintf(fp, "%f ", m_pfMat[i*4+j]);
	    }
	    fprintf(fp, "\n");
	}
    }


    fprintf(fp, "element vertex %d\n", m_ulNVerts);
    fprintf(fp, "property float x\n");
    fprintf(fp, "property float y\n");
    fprintf(fp, "property float z\n");
    fprintf(fp, "property float confidence\n");
    fprintf(fp, "property float intensity\n");
    fprintf(fp, "element face %d\n", m_ulNFaces);
    fprintf(fp, "property list uchar int vertex_indices\n");
    fprintf(fp, "end_header\n");
    for(int i = 0; i < m_ulNVerts; i++){
	fprintf(fp, "%f %f %f %f %f\n",
		m_pVerts[i].pos.fx*fScale, m_pVerts[i].pos.fy*fScale, m_pVerts[i].pos.fz*fScale, m_pVerts[i].Y, m_pVerts[i].Y);
    }
    for(int i = 0; i < m_ulNFaces; i++){
	fprintf(fp, "3 %d %d %d\n",
		m_pFaces[i].nF, m_pFaces[i].nS, m_pFaces[i].nT);
    }
    fclose(fp);
    
    return 0;
}

int C3DFile::WriteX(const char* pcFName, float fScale, bool bFlip, bool bAscii, bool bLittle)
{
    int i;
    std::ofstream ofsOut(pcFName, std::ios::out);
    ofsOut.setf(std::ios::showpoint);
    ofsOut << "xof 0303txt 0032\n"
	   << "Header {\n"
	   << " 1;\n"
	   << " 0;\n"
	   << " 1;\n"
	   << "}\n"
	   << "\n"
	   << "Mesh {\n";
    ofsOut << m_ulNVerts << ";\n";
    for(i = 0; i < m_ulNVerts-1; i++){
	ofsOut << m_pVerts[i].pos.fx*fScale << ";" << m_pVerts[i].pos.fy*fScale << 
	    ";" << m_pVerts[i].pos.fz*fScale << ";,\n";
    }
    ofsOut << m_pVerts[m_ulNVerts-1].pos.fx*fScale << ";" << m_pVerts[m_ulNVerts-1].pos.fy*fScale  
	   << ";" << m_pVerts[m_ulNVerts-1].pos.fz*fScale << ";;\n";
    ofsOut << m_ulNFaces << ";\n";
    for(i = 0; i < m_ulNFaces-1; i++){
	if(bFlip){
	    ofsOut << "3;" << m_pFaces[i].nF << "," << m_pFaces[i].nT << "," << m_pFaces[i].nS << ";,\n";
	}else{
	    ofsOut << "3;" << m_pFaces[i].nF << "," << m_pFaces[i].nS << "," << m_pFaces[i].nT << ";,\n";
	}
    }
    if(bFlip){
	ofsOut << "3;" << m_pFaces[m_ulNFaces-1].nF << "," << m_pFaces[m_ulNFaces-1].nT << ","
	       << m_pFaces[m_ulNFaces-1].nS << ";;\n";
    }else{
	ofsOut << "3;" << m_pFaces[m_ulNFaces-1].nF << "," << m_pFaces[m_ulNFaces-1].nS << ","
	       << m_pFaces[m_ulNFaces-1].nT << ";;\n";
    }
    
    ofsOut << "MeshNormals {\n";
    ofsOut << m_ulNVerts << ";\n";
    for(i = 0; i < m_ulNVerts-1; i++){
	ofsOut << m_pVerts[i].normal.fx << ";" << m_pVerts[i].normal.fy << 
	    ";" << m_pVerts[i].normal.fz << ";,\n";
    }
    ofsOut << m_pVerts[m_ulNVerts-1].normal.fx << ";" << m_pVerts[m_ulNVerts-1].pos.fy 
	   << ";" << m_pVerts[m_ulNVerts-1].normal.fz << ";;\n";
    ofsOut << m_ulNFaces << ";\n";
    for(i = 0; i < m_ulNFaces-1; i++){
	if(bFlip){
	    ofsOut << "3;" << m_pFaces[i].nF << "," << m_pFaces[i].nT << "," << m_pFaces[i].nS << ";,\n";
	}else{
	    ofsOut << "3;" << m_pFaces[i].nF << "," << m_pFaces[i].nS << "," << m_pFaces[i].nT << ";,\n";
	}
    }
    if(bFlip){
	ofsOut << "3;" << m_pFaces[m_ulNFaces-1].nF << "," << m_pFaces[m_ulNFaces-1].nT << ","
	       << m_pFaces[m_ulNFaces-1].nS << ";;\n";
    }else{
	ofsOut << "3;" << m_pFaces[m_ulNFaces-1].nF << "," << m_pFaces[m_ulNFaces-1].nS << ","
	       << m_pFaces[m_ulNFaces-1].nT << ";;\n";
    }
    
    ofsOut << "}\n";
    ofsOut << "}\n";
    
    return 0;
}

int C3DFile::ReadPly(const char* pcFName, float fScale, bool bReverse)
{
    std::ifstream ifPlt(pcFName, std::ios::in | std::ios::binary);
    if(ifPlt.bad() != 0){
	std::cerr << "Could not open PLY." << std::endl;
	return 1;
    }
    
    char strbuff[256];
    char dummy1[64], dummy2[64];
    bool bIsBinary(false);
    // read header
    while(1){
	ifPlt.getline(strbuff, 256);
	if(strstr(strbuff, "format binary_big_endian 1.0") != NULL){
	    std::cerr << "Cannot read big endians" << std::endl;
	    ifPlt.close();
	    return 1;
	}
	if(strstr(strbuff, "format binary_little_endian 1.0") != NULL){
	    bIsBinary = true;
	}
	if(strstr(strbuff, "format ascii 1.0") != NULL){
	    bIsBinary = false;
	}
	if(strstr(strbuff, "element vertex") != NULL){
	    sscanf(strbuff, "%s %s %d", &dummy1, &dummy2, &m_ulNVerts);
	}
	if(strstr(strbuff, "element face") != NULL){
	    sscanf(strbuff, "%s %s %d", &dummy1, &dummy2, &m_ulNFaces);
	}
	if(strstr(strbuff, "end_header") != NULL){
	    break;
	}
    }
    std::cout << '\t' << pcFName << std::endl;
    std::cout << "\tVerts: " <<  m_ulNVerts << " Faces: " << m_ulNFaces << std::endl;
    if(m_pVerts != NULL) delete [] m_pVerts;
    if(m_pFaces != NULL) delete [] m_pFaces; 
    m_pVerts = new VRTX[m_ulNVerts];
    m_pFaces = new TRNGL[m_ulNFaces];
    unsigned long ulcnt(0);
    float fdummy;
    // read vertices
    float fx, fy, fz, fY;
    while(ulcnt < m_ulNVerts){
	if(bIsBinary){
	    ifPlt.read((char *)&(fx), sizeof(float));
	    ifPlt.read((char *)&(fy), sizeof(float));
	    ifPlt.read((char *)&(fz), sizeof(float));
	    ifPlt.read((char *)&(fdummy), sizeof(float));
	    ifPlt.read((char *)&(fY), sizeof(float));
	}else{
	    ifPlt.getline(strbuff, 256);
	    sscanf(strbuff, "%f %f %f %f %f",
		   &fx, &fy, &fz,
		   &fdummy, &fY);
	}
	
	m_pVerts[ulcnt].pos.fx = fx;
	m_pVerts[ulcnt].pos.fy = fy;
	m_pVerts[ulcnt].pos.fz = fz;
	m_pVerts[ulcnt].Y = fY;
	m_pVerts[ulcnt].boundary = false;
	
	m_pVerts[ulcnt].pos.fx *= fScale;
	m_pVerts[ulcnt].pos.fy *= fScale;
	m_pVerts[ulcnt].pos.fz *= fScale;

	ulcnt++;
    }
    // read faces
    ulcnt = 0;
    int ndummy;
    int nF, nS, nT;
    while(ulcnt < m_ulNFaces){
	if(bIsBinary){
	    ifPlt.read((char *)&(ndummy), sizeof(char));
	    ifPlt.read((char *)&(nF), sizeof(int));
	    if(bReverse){
		ifPlt.read((char *)&(nT), sizeof(int));
		ifPlt.read((char *)&(nS), sizeof(int));
		
	    }else{
		ifPlt.read((char *)&(nS), sizeof(int));
		ifPlt.read((char *)&(nT), sizeof(int));
	    }
	}else{
	    ifPlt.getline(strbuff, 256);
	    if(bReverse){
		sscanf(strbuff, "%d %d %d %d",
		       &ndummy, &nF, &nT, &nS);
	    }else{
		sscanf(strbuff, "%d %d %d %d",
		       &ndummy, &nF, &nS, &nT);
	    }
	}
	m_pFaces[ulcnt].nF = nF;
	m_pFaces[ulcnt].nS = nS;
	m_pFaces[ulcnt].nT = nT;

	ulcnt++;
    }
    
    ifPlt.close();
    
    return 0;
}


void C3DFile::Transform(const float* pfM, bool bT)
{
    gsl_matrix* pgm = gsl_matrix_alloc(4, 4);
    gsl_matrix* pgmR = gsl_matrix_alloc(4, 4);
    Utils::gsl_matrix4x4_from_pfloat(pgm, pfM, bT);
    gsl_matrix_memcpy(pgmR, pgm);
    for(int i = 0; i < 3; i++)
	gsl_matrix_set(pgmR, i, 3, 0.0);

    gsl_vector* pgv = gsl_vector_alloc(4);
    for(unsigned long ul = 0; ul < m_ulNVerts; ul++){
	gsl_vector_set(pgv, 0, m_pVerts[ul].pos.fx);
	gsl_vector_set(pgv, 1, m_pVerts[ul].pos.fy);
	gsl_vector_set(pgv, 2, m_pVerts[ul].pos.fz);
	gsl_vector_set(pgv, 3, 1.0);
	Utils::gsl_matrix_vector_mult(pgm, pgv, pgv);
	m_pVerts[ul].pos.fx = gsl_vector_get(pgv, 0);
	m_pVerts[ul].pos.fy = gsl_vector_get(pgv, 1);
	m_pVerts[ul].pos.fz = gsl_vector_get(pgv, 2);

	gsl_vector_set(pgv, 0, m_pVerts[ul].normal.fx);
	gsl_vector_set(pgv, 1, m_pVerts[ul].normal.fy);
	gsl_vector_set(pgv, 2, m_pVerts[ul].normal.fz);
	gsl_vector_set(pgv, 3, 1.0);
	Utils::gsl_matrix_vector_mult(pgmR, pgv, pgv);
	m_pVerts[ul].normal.fx = gsl_vector_get(pgv, 0);
	m_pVerts[ul].normal.fy = gsl_vector_get(pgv, 1);
	m_pVerts[ul].normal.fz = gsl_vector_get(pgv, 2);
    }

    gsl_matrix_free(pgm);
    gsl_matrix_free(pgmR);
    gsl_vector_free(pgv);
}
    
	



int C3DFile::CheckBoundary()
{
    int *pnVCnt;
    pnVCnt = new int[m_ulNVerts];
    for(unsigned long ul = 0; ul < m_ulNVerts; ul++){
	pnVCnt[ul] = 0;
    }
    for(unsigned long ul = 0; ul < m_ulNFaces; ul++){
	pnVCnt[m_pFaces[ul].nF]++;
	pnVCnt[m_pFaces[ul].nS]++;
	pnVCnt[m_pFaces[ul].nT]++;
    }
    for(unsigned long ul = 0; ul < m_ulNVerts; ul++){
	if(pnVCnt[ul] != 6) m_pVerts[ul].boundary = true;
    }
    
    delete [] pnVCnt;
    
    return 0;
}


int C3DFile::ComputeNormals()
{
    int *pnVCnt;
    pnVCnt = new int[m_ulNVerts];
    for(unsigned long ul = 0; ul < m_ulNVerts; ul++){
	pnVCnt[ul] = 0;
	m_pVerts[ul].normal.fx = m_pVerts[ul].normal.fy = m_pVerts[ul].normal.fz = 0.0;
    }
    gsl_vector* v01 = gsl_vector_alloc(3);
    gsl_vector* v20 = gsl_vector_alloc(3);
    gsl_vector* vtmp = gsl_vector_alloc(3);

    for(unsigned long ul = 0; ul < m_ulNFaces; ul++){		
	gsl_vector_set(v01, 0, m_pVerts[m_pFaces[ul].nS].pos.fx-m_pVerts[m_pFaces[ul].nF].pos.fx);
	gsl_vector_set(v01, 1, m_pVerts[m_pFaces[ul].nS].pos.fy-m_pVerts[m_pFaces[ul].nF].pos.fy);
	gsl_vector_set(v01, 2, m_pVerts[m_pFaces[ul].nS].pos.fz-m_pVerts[m_pFaces[ul].nF].pos.fz);

	gsl_vector_set(v20, 0, m_pVerts[m_pFaces[ul].nF].pos.fx-m_pVerts[m_pFaces[ul].nT].pos.fx);
	gsl_vector_set(v20, 1, m_pVerts[m_pFaces[ul].nF].pos.fy-m_pVerts[m_pFaces[ul].nT].pos.fy);
	gsl_vector_set(v20, 2, m_pVerts[m_pFaces[ul].nF].pos.fz-m_pVerts[m_pFaces[ul].nT].pos.fz);
	
	Utils::gsl_vector3_cross(v01, v20, vtmp);
	Utils::gsl_vector_normalize(vtmp);

	m_pFaces[ul].normal.fx = gsl_vector_get(vtmp, 0); 
	m_pFaces[ul].normal.fy = gsl_vector_get(vtmp, 1); 
	m_pFaces[ul].normal.fz = gsl_vector_get(vtmp, 2);
	
	pnVCnt[m_pFaces[ul].nF]++; 
	m_pVerts[m_pFaces[ul].nF].normal.fx += m_pFaces[ul].normal.fx;
	m_pVerts[m_pFaces[ul].nF].normal.fy += m_pFaces[ul].normal.fy;
	m_pVerts[m_pFaces[ul].nF].normal.fz += m_pFaces[ul].normal.fz;
	pnVCnt[m_pFaces[ul].nS]++;
	m_pVerts[m_pFaces[ul].nS].normal.fx += m_pFaces[ul].normal.fx;
	m_pVerts[m_pFaces[ul].nS].normal.fy += m_pFaces[ul].normal.fy;
	m_pVerts[m_pFaces[ul].nS].normal.fz += m_pFaces[ul].normal.fz;
	pnVCnt[m_pFaces[ul].nT]++;
	m_pVerts[m_pFaces[ul].nT].normal.fx += m_pFaces[ul].normal.fx;
	m_pVerts[m_pFaces[ul].nT].normal.fy += m_pFaces[ul].normal.fy;
	m_pVerts[m_pFaces[ul].nT].normal.fz += m_pFaces[ul].normal.fz;
    }

    // get the average normal, too
    gsl_vector* vavg = gsl_vector_alloc(3);
    gsl_vector_set_zero(vavg);
    
    for(unsigned long ul = 0; ul < m_ulNVerts; ul++){
	if(pnVCnt[ul] != 6) m_pVerts[ul].boundary = true;
	else m_pVerts[ul].boundary = false;
	if(pnVCnt[ul] == 0){
	    gsl_vector_set_zero(vtmp);
	}else{
	    gsl_vector_set(vtmp, 0, m_pVerts[ul].normal.fx / pnVCnt[ul]);
	    gsl_vector_set(vtmp, 1, m_pVerts[ul].normal.fy / pnVCnt[ul]);
	    gsl_vector_set(vtmp, 2, m_pVerts[ul].normal.fz / pnVCnt[ul]);
	    Utils::gsl_vector_normalize(vtmp);
	}
	m_pVerts[ul].normal.fx = gsl_vector_get(vtmp, 0);
	m_pVerts[ul].normal.fy = gsl_vector_get(vtmp, 1);
	m_pVerts[ul].normal.fz = gsl_vector_get(vtmp, 2);
	gsl_vector_add(vavg, vtmp);
    }

    gsl_vector_scale(vavg, (double)m_ulNVerts);
    Utils::gsl_vector_normalize(vavg);
    m_d3AvgN.dx = gsl_vector_get(vavg, 0);
    m_d3AvgN.dy = gsl_vector_get(vavg, 1);
    m_d3AvgN.dz = gsl_vector_get(vavg, 2);
    
    
    gsl_vector_free(vavg);
    gsl_vector_free(v01);
    gsl_vector_free(v20);
    gsl_vector_free(vtmp);
    delete [] pnVCnt;
    return 0;
}

int C3DFile::ComputeCentroid()
{
    m_d3Cntrd.dx = m_d3Cntrd.dy = m_d3Cntrd.dz = 0.0;
    for(unsigned long ul = 0; ul < m_ulNVerts; ul++){
	m_d3Cntrd.dx += m_pVerts[ul].pos.fx;
	m_d3Cntrd.dy += m_pVerts[ul].pos.fy;
	m_d3Cntrd.dz += m_pVerts[ul].pos.fz;
    }
    m_d3Cntrd.dx /= (double)m_ulNVerts;
    m_d3Cntrd.dy /= (double)m_ulNVerts;
    m_d3Cntrd.dz /= (double)m_ulNVerts;

    return 0;
}

int C3DFile::ComputeMinMax()
{
    float fMinX(100000.0), fMinY(100000.0), fMinZ(100000.0), fMaxX(-100000.0), fMaxY(-100000.0), fMaxZ(-100000.0);
    
    for(unsigned long ul = 0; ul < m_ulNVerts; ul++){
	if(m_pVerts[ul].pos.fx < fMinX) fMinX = m_pVerts[ul].pos.fx;
	if(m_pVerts[ul].pos.fx > fMaxX) fMaxX = m_pVerts[ul].pos.fx;
	if(m_pVerts[ul].pos.fy < fMinY) fMinY = m_pVerts[ul].pos.fy;
	if(m_pVerts[ul].pos.fy > fMaxY) fMaxY = m_pVerts[ul].pos.fy;
	if(m_pVerts[ul].pos.fz < fMinZ) fMinZ = m_pVerts[ul].pos.fz;
	if(m_pVerts[ul].pos.fz > fMaxZ) fMaxZ = m_pVerts[ul].pos.fz;
    }
    
    m_fMinX = fMinX; 
    m_fMinY = fMinY;
    m_fMinZ = fMinZ;
    m_fMaxX = fMaxX;
    m_fMaxY = fMaxY;
    m_fMaxZ = fMaxZ;

    float fX, fY, fZ;
    fX = fabs(m_fMaxX-m_fMinX);
    fY = fabs(m_fMaxY-m_fMinY);
    fZ = fabs(m_fMaxZ-m_fMinZ);
    
    m_fLong = fX > fY ? fX : fY;
    m_fLong = m_fLong > fZ ? m_fLong : fZ;

	return 0;
}

    
    
    

int C3DFile::ReadFile(const char* pcFName, float fScale, bool bReverse)
{
    // set file name
    std::string strFName(pcFName);
    m_pcFName = new char[strFName.length()+1];
    strcpy(m_pcFName, strFName.c_str());
    int nPos;
#ifndef W32
    nPos = strFName.find_last_of("/");
#endif 
#ifdef W32
    nPos = strFName.find_last_of('\\');
#endif // W32
    
    std::string strSFName = strFName.substr(nPos+1, strFName.length());
    m_pcShortFName = new char[strSFName.length()+1];
    strcpy(m_pcShortFName, strSFName.c_str());
    // check the extension
    nPos = strFName.find_last_of(".");
    if(nPos == -1){
	std::cerr << "Could not identify the file type of " << strFName << std::endl;
	return 1;
    }
    std::string strExt = strFName.substr(nPos+1, strFName.length());
    if(strExt == "ply" || strExt == "PLY"){
	ReadPly(pcFName, fScale, bReverse);
    }else{
	std::cerr << "cannot handle " << strExt << " files." << std::endl;
	return 1;
    }
    
    ComputeNormals();
    CheckBoundary();
    ComputeCentroid();
    ComputeMinMax();
    return 0;
}
