/**************************************************************************
 *                                                                        *
 *                       This file is part of PSReg                       *
 *                                                                        *
 *          Copyright (C) 2002 Ko Nishino. All Rights Reserved.           *
 *                                                                        *
 *  Permission to use this software for educational and research purposes *
 * is hereby granted without fee. All other rights, including the rights  *
 * to distribute, modify, or sell the software, are explicitly reserved   *
 * by the author. Please contact the author if you wish to make any       *
 * modifications or additions to the software.                            *
 *                                                                        *
 *  This software is provided "as is" and without warranty of any kind,   *
 * express, implied or otherwise, including without limitation, any       *
 * warranty of merchantability or fitness for a particular purpose.       *
 *                                                                        *
 **************************************************************************/
/*  
 *  KDTree.cpp
 *          Ko Nishino 
 *               kon@cvl.iis.u-tokyo.ac.jp
 *               http://www.cvl.iis.u-tokyo.ac.jp/~kon/
 */
/*
 * Adapted from Mark Wheeler's kdtree implementation.
 */

#include <stdlib.h>
#include <iostream>
#include "KDefs.h"
#include "KDTreeStruct.h"
#include "KDTree.h"

CKDTree::CKDTree()
{

}

CKDTree::~CKDTree()
{
	/*
	delete [] m_pContents;
	// TODO: USE freeKNNData
	for(int i = 0; i < m_nNRoots; i++){
		delete [] KDNNRoot[i];
	}
	delete [] KDNNRoot;
	*/
}

/* 
 * kd-tree.c 
 * 
 * Created 5-27-92 by Mark Wheeler at Carnegie Mellon University, 
 *                    mdwheel@cs.cmu.edu
 * 
 * 
 * code to build a kd-tree which 
 * stores a set of n k-dimensional vectors
 * with a uniform dissimilarity metric
 * 
 * Algorithm taken from:
 * "An Algorithm for Finding Best Matches in Logarithmic Expected Time"
 *        by Jerome H. Friedman, Jon Bentley and Raphael Finkel
 * found in "ACM Transactions on Mathematical Software"
 *             VOLUME 3 No. 3 1977, pp. 209-226
 * 
 * Must be compiled with librecipes.a (Numerical Recipes)
 * it calls mdian2() to compute the median of a vector
 */
/* 
 * MakeKDTree
 * top level call to build a kd-tree of indices into the array data
 * takes:
 *  root -- a pointer to an allocated KDtree node
 *  data -- an (n,k) array of k-d data
 *  k -- the dimension of the data vectors
 *  n -- the number of data vectors
 * creates a kd-tree pointed to by root
 * the kd-tree leaves contain indices which can be used to index the original
 * data vector
 * this saves much space especially if k is large
 */
void CKDTree::MakeKDTree(BUCKET *root, float data[], int k, int n)
{
    int i, *index;

    index = (int *) malloc((unsigned)sizeof(int)*n);
    
    for( i = 0; i < n; i++) index[i] = i;
    
    BuildKDTree(root,data,index,k,n);

    free((char *) index);

}

/* 
 * BuildKDTree --- algorithm from Friedman, Bentley and Finkel
 * data is an (*,k) vector of floats
 * and index (n) vector of ints which index into elements of data
 */

void CKDTree::BuildKDTree(BUCKET *result,float data[],
			  int index[], int k, int n)
{
    int i, j, *rd, *ld, r, l, count, K, same;
    int d, *left_index, *right_index;
    float p, maxspread, spread_estimate;

  
    if (n <= BUCKET_SIZE) {
	result->type = TERMINAL;
	result->dimension = 0;
	result->median = 0.0;
	result->left = NULL;
	result->right = NULL;
	
	result->index = (int *) malloc((unsigned) sizeof(int)*n);
	if(result->index == NULL){ 
	    std::cerr << "Could not malloc in BuildKDTree" << std::endl;
	    return;
	}
	
	count = 0;
	for(i=0; i < n; ++i) { 
	    /* check for duplicates here - ow this can mess up any application
	     * that assumes 1 entry per location */
	    j = 0; same = 0;
	    while (!same && j < i) {
		same = 1; K = 0;
		while(same && K < k) {
		    same = (data[index[i]*k + K] == data[index[j]*k + K]);
		    ++K;
		}
		++j;
	    }
	    if (!same) {
		result->index[count] = index[i];
		++count;
	    }

	}
	result->n = count;
    }
    else {
	maxspread = 0;
	d = 0;    
	for(j=0; j < k; ++j) {
	    spread_estimate =  KDSpreadEstimate(j,data,index,k,n);
	    if (spread_estimate > maxspread) {
		maxspread = spread_estimate;
		d = j;
	    }
	}
	p = KDMedian(d,data,index,k,n);
	
	left_index = (int *) malloc((unsigned) sizeof(int)*n);
	if(left_index == NULL){ 
	    std::cerr << "Could not malloc in BuildKDTree" << std::endl;
	    return;
	}
	right_index = (int *) malloc((unsigned) sizeof(int)*n);
	if(right_index == NULL){ 
	    std::cerr << "Could not malloc in BuildKDTree" << std::endl;
	    return;
	}
	

	ld = left_index;
	rd = right_index;
	r = 0;     l = 0;
	for(i=0; i < n; ++i)     
	    if (data[index[i]*k + d] <= p) {
		++l;
		*ld++ = index[i];
	    }
	    else {
		++r;
		*rd++ = index[i];
	    }
	
	if ((r == n) || (l == n)) {
	    /*
	     * we have identical elements: store as one element in a terminal node
	     */
	    result->type = TERMINAL;
	    result->dimension = 0;
	    result->median = 0.0;
	    result->left = NULL;
	    result->right = NULL;
	    
	    result->index = (int *) malloc((unsigned) sizeof(int));
	    if(result->index == NULL){
		std::cerr << "Could not malloc in BuildKDTree" << std::endl;
		return;
	    }
	    
	    result->n = 1;
	    result->index[0] = index[0];
	    /* was:
	       fprintf(stderr,
	       "Fatal error: There are %d equivalent entries in k-d tree\n",n);
	       exit(-1);
	    */
	} 
	else {
	    result->left = (BUCKET *) malloc((unsigned) sizeof(BUCKET));
	    if(result->left == NULL){ 
		std::cerr << "Could not malloc in BuildKDTree" << std::endl;
		return;
	    }
	    result->right = (BUCKET *) malloc((unsigned) sizeof(BUCKET));
	    if(result->right == NULL){ 
		std::cerr << "Could not malloc in BuildKDTree" << std::endl;
		return;
	    }
	    
	    BuildKDTree(result->left,data,left_index,k,l);
	    BuildKDTree(result->right,data,right_index,k,r);
	    result->dimension = d;
	    result->median = p;
	    /* zero out other fields */
	    result->type = NONTERMINAL;
	    result->index = NULL;
	    result->n = 0;
	}
	free((char *) left_index);
	free((char *) right_index);
    }
    
}

/* KDMedian
 * return the median of the dth dimension of the vector of data
 * TODO: This function requires mdian3 from Numerical Recipe, which makes the code ugly.
 */
float CKDTree::KDMedian(int d, float *data, 
			int index[],int k,int n)
{
    float *ddata;
    float result;
    int i;
    
    ddata = (float *) malloc((unsigned) sizeof(float)*(n+1));
    
    for(i=0; i < n; ++i)  
	ddata[i+1] = data[index[i]*k + d];
    
    /* code for finding median of a vector of floats 
     * --- from Numerical Recipes 
     * note NR uses vectors from ddata[1] ... ddata[n]
     */
    mdian3(ddata,n,&result);
    free((char *) ddata);
    
    return(result);
}

/* From Numerical Recipe
 * mdian3.c
 * modified to balance the sides
 */
void CKDTree::mdian3(float x[],int n,float *xmed)
{
    int n2,nhi,nlo;
    
    sort(n,x);
    n2 = (n % 2) ? n/2 + 1 : n/2;
    *xmed = x[n2];
    
    /* when the real median is also the maximum our rule if data <= median
     * will not split the data
     * thus we choose a median just below the real median but above the
     * next lowest data 
     */
    if(*xmed < x[n2+1]) return; /* perfect split */
    
    nhi = n2 + 1;
    while((*xmed == x[nhi]) && (nhi < n)) ++nhi;
    
    nlo = n2 - 1;
    while((*xmed == x[nlo]) && (nlo > 1)) --nlo;
    
    if (n - 2*nlo <= 2*nhi - n) 
	*xmed = x[nlo];
    else *xmed = x[nhi];
}

// From Numerical Recipe
void CKDTree::sort(int n, float ra[])
{
    int l,j,ir,i;
    float rra;
    
    l=(n >> 1)+1;
    ir=n;
    for (;;) {
	if (l > 1)
	    rra=ra[--l];
	else {
	    rra=ra[ir];
	    ra[ir]=ra[1];
	    if (--ir == 1) {
		ra[1]=rra;
		return;
	    }
	}
	i=l;
	j=l << 1;
	while (j <= ir) {
	    if (j < ir && ra[j] < ra[j+1]) ++j;
	    if (rra < ra[j]) {
		ra[i]=ra[j];
		j += (i=j);
	    }
	    else j=ir+1;
	}
	ra[i]=rra;
    }
}


/* KDSpreadEstimate
 * return the magnitude of the range of values of the dth 
 * dimension of the vector of data
 */
float CKDTree::KDSpreadEstimate(int d,float *data, int index[], int k,int n)
{
    float sum,mean;
    int i;
    
    sum = 0.0;
    for(i = 0; i < n; ++i) 
	sum += data[index[i]*k+d];
    
    mean = sum/(float) n;
    sum = 0.0;
    for(i = 0; i < n; ++i) 
	sum += (data[index[i]*k+d] - mean)*(data[index[i]*k+d] - mean);
    
    return(sum/(float) n);
}

/* KDTreeSearchNode
 */
BUCKET_PTR CKDTree::KDTreeSearchNode(BUCKET *node, float key[], int k)
{
    if(node->type == TERMINAL){
	/* for each element in the terminal bucket
	 * compute the dissimilarity and insert into the queue
	 */
	return(node);
    }else{
	if(key[node->dimension] <= node->median){
	    return(KDTreeSearchNode(node->left,key,k));
	}else{
	    return(KDTreeSearchNode(node->right,key,k));
	}
    }
}

/* KDTreeInsertItem
 */
void CKDTree::KDTreeInsertItem(float data[], BUCKET *root, float key[], int index, int k)
{
    int i, *new_index_list;
    int index_list[BUCKET_SIZE+1];
    BUCKET *node;
    
    node = KDTreeSearchNode(root,key,k);
    
    if(node->n == BUCKET_SIZE){
	/* add two buckets to this node to add the new point */
	for(i = 0; i < BUCKET_SIZE+1; ++i)
	    index_list[i] = node->index[i];
	index_list[BUCKET_SIZE] = index;
	
	BuildKDTree(node,data,index_list,k,BUCKET_SIZE+1);
    }else{ /* room to add this to index list */
	new_index_list = (int *) malloc((unsigned) sizeof(int)*(node->n + 1));
	for(i = 0; i < node->n; ++i)
	    new_index_list[i] = node->index[i];
	new_index_list[node->n] = index;
	free((char *) node->index);
	node->index = new_index_list;
	node->n = node->n + 1;
    }
}

/* KDTreeRemoveItem
 */
void CKDTree::KDTreeRemoveItem(BUCKET *root, float key[], int index, int k)
{
    int i;
    BUCKET *node;
    
    node = KDTreeSearchNode(root,key,k);
    
    if(node->n == 0) 
	return;
    else{
	i = 0;
	while((i < node->n) && (node->index[i] != index))
	    ++i;
	if(node->index[i] == index){
	    while(i+1 < node->n){
		node->index[i] = node->index[i+1];
		++i;
	    }
	    --(node->n);
	}
    }
}

/* 
 * k-nearest-neighbors.c 
 * 
 * Created 5-27-92 by Mark Wheeler at Carnegie Mellon University, 
 *                    mdwheel@cs.cmu.edu
 * 
 * Code to do k-nearest neighbors search of k-dimensional vectors
 * with a uniform dissimilarity metric utilizing kd-trees
 * 
 * 
 * Algorithm taken from:
 * "An Algorithm for Finding Best Matches in Logarithmic Expected Time"
 *        by Jerome H. Friedman, Jon Bentley and Raphael Finkel
 * found in "ACM Transactions on Mathematical Software"
 *             VOLUME 3 No. 3 1977, pp. 209-226
 *
 * uses code in kd-tree.c to build the optimized kd-tree for efficient search
 *
 * Must be compiled with librecipes.a (Numerical Recipes)
 * it uses mdian2() for finding the median of a vector
 *
 *
 * Modified 5-15-93 by David Simon - Optimized for 3d trees. It is assumed
 * that add data points are 3 dimensional. Also added register variables. 
 *
 */

/*
 * NOTE: this appears to be faster than the above, despite the claims of
 * Sproull.  
 */
#define ComputeDissimilarity(x, y, k, dis) {  float t; int i; (dis) = 0; for(i=0; i < k; ++i) { CoordinateDistance(i,x,y,t); (dis) += t; } }


/* 
 * top level m-nearest-neighbor search routine (initializes variables required
 * for the recursive search routine):
 * takes:
 *   root -- a pointer to a kd-tree node
 *   data -- an array of k-dimensional points
 *   k    -- dimension of points
 *   m    -- number of neighbors to find
 * returns:
 *   pqd  -- m-dimensional vector of squared distances for the m-nearest 
 *           neighbors in increasing order of distance
 *   pqr  -- m-dimensional vector of indices (of the data array) for the 
 *           m-nearest neighbors in increasing order of distance
 */
void CKDTree::KDNNSearch(BUCKET *root, float data[], float key[], float pqd[],
			 int pqr[], int k, int m, bool bCheck, int n, 
			 int nstop, int& nresult)
{
    int i;
    int ncnt(0);
    
    for(i = 0; i < m; ++i) pqd[i] = MAXFLOAT;
    for(i = 0; i < m; ++i) pqr[i] = -1;
    KDNNSearchNode(root,data,key,pqd,pqr,k,m, n, ncnt, nstop, nresult);
    if(nresult == 0){ // aborted

    }else{
    }
}


/* KDNNSearchNode -- algorithm from Friedman, Bentley and Finkel
 * recursive nearest neighbor search routine -- to mimick tail recursion
 * returns TRUE when done to exit previous calls
 */
void CKDTree::KDNNSearchNode(BUCKET *node, float data[], float key[], float pqd[],
			     int pqr[], int k, int m, int n, int& ncnt, int nstop, int& nresult)
{
    int i, j, d;
    float dist, temp, p;
    float dp, dp2;
    float *dataptr;
    if(ncnt == nstop){
	nresult = 0;
	return;
    }else{
	ncnt++;
    }
    

    if(node->type == TERMINAL){
	/* for each element in the terminal bucket
	 * compute the dissimilarity and insert into the queue
	 */
	for(i = 0; i < node->n; ++i){
	    dataptr = &data[node->index[i]*k];
	    ComputeDissimilarity(key, dataptr, k, dist);
	    /* if this element is closer than any of the m closest
	     * insert it into the queue
	     */
	    if(dist < pqd[m-1]){
		j = m-1;
		pqd[j] = dist;
		pqr[j] = node->index[i];
		while((j-1 >= 0) && (pqd[j] < pqd[j-1])){
		    temp = pqd[j-1]; pqd[j-1] = pqd[j]; pqd[j] = temp;
		    temp = pqr[j-1]; pqr[j-1] = pqr[j]; pqr[j] = (int)temp;
		    --j;
		}
	    }
	}
	nresult = ncnt;
    }else{
	d = node->dimension;
	p = node->median;
	dp = key[d] - p;
	dp2 = dp * dp;
	
	if(dp <= 0){ /* was < , but <= goes left in kd-tree.c(shouldn't matter)*/
	    KDNNSearchNode(node->left,data,key,pqd,pqr,k,m,n, ncnt, nstop, nresult);
	    if(dp2 < pqd[m-1]){
		KDNNSearchNode(node->right,data,key,pqd,pqr,k,m,n, ncnt, nstop, nresult);
	    }
	}else{
	    KDNNSearchNode(node->right,data,key,pqd,pqr,k,m,n, ncnt, nstop, nresult);
	    if(dp2 < pqd[m-1]){
		KDNNSearchNode(node->left,data,key,pqd,pqr,k,m,n, ncnt, nstop, nresult);
	    }
	}
    }
}


/* KNNSearch provides an interface for doing multiple nearest neighbor 
 * searches while hiding the kd-tree structure from the user.
 * A static variable is used to make sure the kdtree is built only once.
 * the interface is:
 *   data -- an array of k-dimensional points
 *   k    -- dimension of points
 *   m    -- number of neighbors to find
 *   n    -- number of vectors in data

 *   i    -- kd tree index, allow multiple simultaneous	trees
 * returns:
 *   pqd  -- m-dimensional vector of squared distances for the m-nearest
 *           neighbors in increasing order of distance
 *   pqr  -- m-dimensional vector of indices (of the data array) for the 
 *           m-nearest neighbors in increasing order of distance
 * it is assumed that pqd and pqr are allocated by the calling program.
 */
/* static variables visible only to this interface */
//static BUCKET *KDNNRoot[200];

void CKDTree::KNNSearch(float data[], float key[], float pqd[],
			int pqr[], int k, int m, int n, int i, bool bCheck, 
			int nstop, int& nresult)
{
    if(KDNNRoot[i] == NULL){
	/* Build KDTree first time only */
	KDNNRoot[i] = (BUCKET *) malloc((unsigned) sizeof(BUCKET));
	MakeKDTree(KDNNRoot[i],data,k,n);
    }
    
    KDNNSearch(KDNNRoot[i], data, key, pqd, pqr, k, m, bCheck, n, nstop, nresult);
}

/* set KDNNFirstTime to TRUE so tree is rebuilt next time */
void CKDTree::NewKNNData(int i)
{

    if(KDNNRoot[i]){
	/* free the tree data structures */
	freeKNNTree(KDNNRoot[i]);
	KDNNRoot[i] = NULL;
    }
}


void CKDTree::freeKNNTree(BUCKET *node)
{
  if(node != NULL){
      switch(node->type){
      case TERMINAL:
	  free((char *) node->index);
	  break;
      case NONTERMINAL:
	  freeKNNTree(node->left);
	  freeKNNTree(node->right);
	  break;
      }
      free((char *) node);
  }
}

void CKDTree::KNNInsertItem(float data[], float key[], int k, int index, int i)
{

    KDTreeInsertItem(data,KDNNRoot[i],key,index,k);
    
}


void CKDTree::KNNRemoveItem(float key[], int k, int index, int i)
{
    
    KDTreeRemoveItem(KDNNRoot[i],key,index,k);
}
