/**************************************************************************
 *                                                                        *
 *                       This file is part of PSReg                       *
 *                                                                        *
 *          Copyright (C) 2002 Ko Nishino. All Rights Reserved.           *
 *                                                                        *
 *  Permission to use this software for educational and research purposes *
 * is hereby granted without fee. All other rights, including the rights  *
 * to distribute, modify, or sell the software, are explicitly reserved   *
 * by the author. Please contact the author if you wish to make any       *
 * modifications or additions to the software.                            *
 *                                                                        *
 *  This software is provided "as is" and without warranty of any kind,   *
 * express, implied or otherwise, including without limitation, any       *
 * warranty of merchantability or fitness for a particular purpose.       *
 *                                                                        *
 **************************************************************************/
/*  
 *  PSReg.cpp
 *          Ko Nishino 
 *               kon@cvl.iis.u-tokyo.ac.jp
 *               http://www.cvl.iis.u-tokyo.ac.jp/~kon/
 */

#include "KDefs.h"
#include <GL/gl.h>
#include "KonStructs.h"
#include "PSReg.h"
#include "Pose.h"
#include "Utils.h"
#include "RegKernel.h"
#include <pthread.h>

void* thrdSetKDTree(void* pvoid);

CPSReg::CPSReg()
{
    m_pkdt = NULL;
}

CPSReg::~CPSReg()
{
    cleanObjects();
}

void CPSReg::addObject(C3DFile* pc3dObj)
{
    SROBJ* psrobj = new SROBJ;
    psrobj->ulNVerts = pc3dObj->GetNVerts();
    psrobj->ulNFaces = pc3dObj->GetNFaces();

    // store vertex coordinates
    psrobj->pfC = new float[3*psrobj->ulNVerts];
    psrobj->pfN = new float[3*psrobj->ulNVerts];
    psrobj->pfI = new float[psrobj->ulNVerts];
    psrobj->pbB = new bool[psrobj->ulNVerts];
    VRTX* pVerts;
    pVerts = pc3dObj->GetVerts();
    for(unsigned long ul = 0; ul < psrobj->ulNVerts; ul++){
	psrobj->pfC[ul*3  ] = (float)pVerts[ul].pos.fx;
	psrobj->pfC[ul*3+1] = (float)pVerts[ul].pos.fy;
	psrobj->pfC[ul*3+2] = (float)pVerts[ul].pos.fz;
	psrobj->pfN[ul*3  ] = (float)pVerts[ul].normal.fx;
	psrobj->pfN[ul*3+1] = (float)pVerts[ul].normal.fy;
	psrobj->pfN[ul*3+2] = (float)pVerts[ul].normal.fz;
	psrobj->pfI[ul]   = pVerts[ul].Y;
	psrobj->pbB[ul]   = pVerts[ul].boundary;
    }

    // store face indices
    psrobj->pulF = new unsigned long[3*psrobj->ulNFaces];
    TRNGL* pFaces;
    pFaces = pc3dObj->GetFaces();
    for(unsigned long ul = 0; ul < psrobj->ulNFaces; ul++){
	psrobj->pulF[ul*3  ] = pFaces[ul].nF;
	psrobj->pulF[ul*3+1] = pFaces[ul].nS;
	psrobj->pulF[ul*3+2] = pFaces[ul].nT;
    }

    pc3dObj->GetAvgNormal(psrobj->pfAvgN);
    pc3dObj->GetCentroid(psrobj->pfCntrd);
    pc3dObj->GetLong(psrobj->fMaxLength);
    
    psrobj->pM = gsl_matrix_alloc(4, 4);
    psrobj->pRecM = gsl_matrix_alloc(4, 4);
    gsl_matrix_set_identity(psrobj->pM);
    gsl_matrix_set_identity(psrobj->pRecM);
    m_vsrObjs.push_back(psrobj);
}
	
int CPSReg::createKDTrees()
{
    if(m_pkdt != NULL){
	std::cerr << "KDTrees are already constructed" << std::endl;
	return 1;
    }

    m_pkdt = new CKDTree;
    m_pkdt->Allocate(m_vsrObjs.size());

    int ncnt(0);
    std::vector<SROBJ*>::iterator itr;
    std::cout << "Making kdtrees " << std::flush;

#ifdef KDNONTHREADS
    for(itr = m_vsrObjs.begin(); itr != m_vsrObjs.end(); itr++){
	// TODO: Implement a sampler
	samplePoints(*itr, 1.0); 
	m_pkdt->AllocRoot(ncnt);
	m_pkdt->MakeKDTree(m_pkdt->GetRoot(ncnt), (*itr)->pfC, 3, (*itr)->ulNVerts);
	m_pkdt->SetContent(ncnt, ncnt, (*itr)->ulNVerts, 3, 
			   (*itr)->pfC, (*itr)->pfN, (*itr)->pfI, (*itr)->pbB);

	ncnt++;
    }
#endif // KDNONTHREADS
    
#ifdef KDTHREADS
    for(itr = m_vsrObjs.begin(); itr != m_vsrObjs.end(); itr++){
	samplePoints(*itr, 1.0);
    }

    argsSKDT** pArgs = new (argsSKDT (*[m_vsrObjs.size()]));
    pthread_t* pthrds = new pthread_t[m_vsrObjs.size()];
    pthread_mutex_t mutex;
    pthread_mutex_init(&mutex, NULL);
    ncnt = 0;
    for(itr = m_vsrObjs.begin(); itr != m_vsrObjs.end(); itr++){
	pArgs[ncnt] = new argsSKDT;
	pArgs[ncnt]->pmutex = &mutex;
	pArgs[ncnt]->nID = ncnt;
	pArgs[ncnt]->pkdt = m_pkdt;
	pArgs[ncnt]->pObj = *itr;

	pthread_create(&(pthrds[ncnt]), NULL, thrdSetKDTree, pArgs[ncnt]);
	ncnt++;
    }
    for(int i = 0; i < m_vsrObjs.size(); i++){
	pthread_join(pthrds[i], NULL);
    }
    
    for(int i = 0; i < m_vsrObjs.size(); i++){
	delete pArgs[i];
    }
    delete [] pArgs;
    delete [] pthrds;

#endif // KDTHREADS


    std::cout << " done" << std::endl;
    
    return 0;
}

void* thrdSetKDTree(void* pvoid)
{
    argsSKDT* pargs = (argsSKDT*)pvoid;
    pthread_mutex_lock(pargs->pmutex);
    pargs->pkdt->AllocRoot(pargs->nID);
    pthread_mutex_unlock(pargs->pmutex);
    pthread_mutex_lock(pargs->pmutex);
    pargs->pkdt->MakeKDTree(pargs->pkdt->GetRoot(pargs->nID), 
			    pargs->pObj->pfC, 3, (int)pargs->pObj->ulNVerts);
    pthread_mutex_unlock(pargs->pmutex);
    pthread_mutex_lock(pargs->pmutex);
    pargs->pkdt->SetContent(pargs->nID, pargs->nID,
			    pargs->pObj->ulNVerts, 3,
			    pargs->pObj->pfC, pargs->pObj->pfN, 
			    pargs->pObj->pfI, pargs->pObj->pbB);
    pthread_mutex_unlock(pargs->pmutex);
    std::cout << "*" << std::flush;
    return NULL;
}
    

int CPSReg::deleteKDTrees()
{
    if(m_pkdt == NULL){
	std::cout << "KDTree not allocated." << std::endl;
	return 1;
    }
    m_pkdt->FreeAll();
    m_pkdt = NULL;
    return 0;
}

/* 
 */
void CPSReg::updateMatrices(std::vector<GLfloat*>* pvmats)
{
    gsl_matrix* pgDM = gsl_matrix_alloc(4, 4);
    GLfloat* pglf;
    for(int i = 0; i < pvmats->size(); i++){ // const didn't allow using iterator
	pglf = (*pvmats)[i];
	Utils::gsl_matrix4x4_from_pfloat(pgDM, (float*)pglf, true);
	Utils::gsl_matrix4x4_mult(pgDM, m_vsrObjs[i]->pRecM, m_vsrObjs[i]->pRecM);
	gsl_matrix_memcpy(m_vsrObjs[i]->pM, pgDM);
    }
    gsl_matrix_free(pgDM);
    
    updateCoords();
    clearPMs();
}

void CPSReg::uploadMatrices(std::vector<GLfloat*>* pvmats)
{
    gsl_matrix* pgM = gsl_matrix_alloc(4, 4);;
    GLfloat* pglf;
    for(int i = 0; i < pvmats->size(); i++){
	pglf = (*pvmats)[i];
	Utils::gsl_matrix4x4_from_pfloat(pgM, (float*)pglf, true); 
	Utils::gsl_matrix4x4_mult(m_vsrObjs[i]->pM, pgM, pgM);
	Utils::gsl_matrix4x4_to_pfloat(pgM, (*pvmats)[i], true);
	gsl_matrix_set_identity(m_vsrObjs[i]->pM);
    }
    gsl_matrix_free(pgM);
}
    

int CPSReg::samplePoints(SROBJ* srobj, float fPrcnt)
{
    if(fPrcnt != 1.0){
	randomSampling(srobj, fPrcnt);
    }else{
	// nop
    }
    
    return 0;
}

int CPSReg::randomSampling(SROBJ* srobj, float fPrcnt)
{
    

    return 0;
}

int CPSReg::runRegistration(const int nNSteps, const int nNLMSteps, 
			    const float fMESigma, const float fNThresh,
			    const float fTDSearch, const float fCGSearch, 
			    const bool bBCheck, const bool bIEval,
			    const int nMultiNP,
			    const bool bPairwise, bool* pbModels,
			    bool** pbScnList)
{
    CPose* pcpPs;
    pcpPs = new CPose[m_vsrObjs.size()];
    // Make sure SRObjs are already updated before calling createKDTrees
    createKDTrees();
    std::cout << "createKDTrees done" << std::endl;
    for(int nid = 0; nid < m_vsrObjs.size(); nid++){
	std::cout << "going " << nid << std::endl;
	if(bPairwise){
	    std::cout << "pairwise" << std::endl;
	    if(pbModels[nid]){
		std::cout << "going: " << nid << std::endl;
		CRegKernel crkernel(m_pkdt, nid, pbScnList[nid],
				    nNSteps, nNLMSteps, fMESigma, fNThresh, 
				    fTDSearch, fCGSearch, bBCheck, bIEval, nMultiNP);
		crkernel.run(pcpPs[nid]);
	    }else{
		std::cout << "skipping: " << nid << std::endl;
	    }
	}else{
	    CRegKernel crkernel(m_pkdt, nid, pbScnList[nid],
				nNSteps, nNLMSteps, fMESigma, fNThresh, 
				(bIEval ? LARGE_INT : fTDSearch),
				(bIEval ? LARGE_INT : fCGSearch),
				bBCheck, bIEval, nMultiNP);				
	    crkernel.run(pcpPs[nid]);
	}
    }

    // update object coords for the next run (important for consistency)
    gsl_matrix* pgm = gsl_matrix_alloc(4, 4);
    for(int nid = 0; nid < m_vsrObjs.size(); nid++){
	pcpPs[nid].checkZero();
	pcpPs[nid].print();
	pcpPs[nid].getMatrix(pgm);
	gsl_matrix_memcpy(m_vsrObjs[nid]->pM, pgm);
	Utils::gsl_matrix4x4_mult(pgm, m_vsrObjs[nid]->pRecM, m_vsrObjs[nid]->pRecM);
    }
    updateCoords();
    std::cout << std::endl;
    std::cout << std::endl;
    gsl_matrix_free(pgm);
    deleteKDTrees();
    delete [] pcpPs;

    return 0; 

}

int CPSReg::dump3DFile(const int nID, C3DFile* pc3df)
{
    assert(nID < m_vsrObjs.size());


    SROBJ* pObj = m_vsrObjs[nID];
    pc3df->SetNVerts(pObj->ulNVerts);
    pc3df->SetNFaces(pObj->ulNFaces);
    pc3df->Alloc();
    std::cout << pc3df->GetNVerts() << " " << pc3df->GetNFaces() << std::endl;
    VRTX* pV = pc3df->GetVerts();
    for(unsigned long ul = 0; ul < pc3df->GetNVerts(); ul++){
	pV[ul].pos.fx = pObj->pfC[ul*3  ];
	pV[ul].pos.fy = pObj->pfC[ul*3+1];
	pV[ul].pos.fz = pObj->pfC[ul*3+2];
	pV[ul].normal.fx = pObj->pfN[ul*3  ];
	pV[ul].normal.fy = pObj->pfN[ul*3+1];
	pV[ul].normal.fz = pObj->pfN[ul*3+2];
	pV[ul].boundary = pObj->pbB[ul];
	pV[ul].Y = pObj->pfI[ul];
    }
    TRNGL* pF = pc3df->GetFaces();
    for(unsigned long ul = 0; ul < pc3df->GetNFaces(); ul++){
	pF[ul].nF = pObj->pulF[ul*3  ];
	pF[ul].nS = pObj->pulF[ul*3+1];
	pF[ul].nT = pObj->pulF[ul*3+2];
    }
    
    float pfMat[16];
    Utils::gsl_matrix4x4_to_pfloat(pObj->pRecM, pfMat, false);
    pc3df->SetMatrix(pfMat, true);
    
    return 0;
}
	
