/**************************************************************************
 *                                                                        *
 *                       This file is part of PSReg                       *
 *                                                                        *
 *          Copyright (C) 2002 Ko Nishino. All Rights Reserved.           *
 *                                                                        *
 *  Permission to use this software for educational and research purposes *
 * is hereby granted without fee. All other rights, including the rights  *
 * to distribute, modify, or sell the software, are explicitly reserved   *
 * by the author. Please contact the author if you wish to make any       *
 * modifications or additions to the software.                            *
 *                                                                        *
 *  This software is provided "as is" and without warranty of any kind,   *
 * express, implied or otherwise, including without limitation, any       *
 * warranty of merchantability or fitness for a particular purpose.       *
 *                                                                        *
 **************************************************************************/
/*  
 *  PSReg.h
 *          Ko Nishino 
 *               kon@cvl.iis.u-tokyo.ac.jp
 *               http://www.cvl.iis.u-tokyo.ac.jp/~kon/
 */

#ifndef _PSREG_H_
#define _PSREG_H_

#include <vector>
#include <gsl/gsl_matrix.h>
#include <pthread.h>
#include "Utils.h"
#include "3DFile.h"
#include "KDTree.h"

typedef struct _SROBJ{
    float* pfC; // Coordinates
    float* pfN; // Normals
    float* pfI; // Intensities
    unsigned long* pulF; // Face indices
    bool* pbB; // Boundary flag
    unsigned long ulNVerts;
    unsigned long ulNFaces;
    float pfAvgN[3];
    float pfCntrd[3];
    float fMaxLength;
    gsl_matrix* pM;
    gsl_matrix* pRecM;
    void getAvgNinGSL(gsl_vector* pgv){
	gsl_vector_set(pgv, 0, pfAvgN[0]);
	gsl_vector_set(pgv, 1, pfAvgN[1]);
	gsl_vector_set(pgv, 2, pfAvgN[2]);
	gsl_vector_set(pgv, 3, 1.0);
	gsl_matrix* pgmR = gsl_matrix_alloc(4, 4);
	gsl_matrix_memcpy(pgmR, pM);
	for(int i = 0; i < 3; i++) gsl_matrix_set(pgmR, i, 3, 0.0);
	Utils::gsl_matrix_vector_mult(pgmR, pgv, pgv);
	gsl_matrix_free(pgmR);
    }
    void getCntrdinGSL(gsl_vector* pgv){
	gsl_vector_set(pgv, 0, pfCntrd[0]);
	gsl_vector_set(pgv, 1, pfCntrd[1]);
	gsl_vector_set(pgv, 2, pfCntrd[2]);
	gsl_vector_set(pgv, 3, 1.0);
	Utils::gsl_matrix_vector_mult(pM, pgv, pgv);
    }
    void getModelinGSL(gsl_matrix* pgm){
	gsl_matrix_memcpy(pgm, pM);
    }
    float getMaxLength(){ return fMaxLength; }
    void print(){
	std::cout << "SROBJ matrix: " << std::endl;
	for(int i = 0; i < 4; i++){
	    std::cout << "\t";
	    for(int j = 0; j < 4; j++){
		std::cout << gsl_matrix_get(pM, i, j) << "\t";
	    }
	    std::cout << std::endl;
	}
    }
    void update(){
#ifdef DTLDEBUG
	std::cout << "updating with:" << std::endl;
	print();
#endif // DTLDEBUG
	gsl_vector* pgv = gsl_vector_alloc(4);
	gsl_vector_set(pgv, 3, 1.0);
	gsl_matrix* pgmR = gsl_matrix_alloc(4, 4);
	gsl_matrix_memcpy(pgmR, pM);
	for(int i = 0; i < 3; i++) gsl_matrix_set(pgmR, i, 3, 0.0);
	for(unsigned long ul = 0; ul < ulNVerts; ul++){
	    gsl_vector_set(pgv, 0, pfC[ul*3  ]); gsl_vector_set(pgv, 1, pfC[ul*3+1]); gsl_vector_set(pgv, 2, pfC[ul*3+2]);
	    Utils::gsl_matrix_vector_mult(pM, pgv, pgv);
	    pfC[ul*3  ] = gsl_vector_get(pgv, 0); pfC[ul*3+1] = gsl_vector_get(pgv, 1); pfC[ul*3+2] = gsl_vector_get(pgv, 2);
	    gsl_vector_set(pgv, 0, pfN[ul*3  ]); gsl_vector_set(pgv, 1, pfN[ul*3+1]); gsl_vector_set(pgv, 2, pfN[ul*3+2]);
	    Utils::gsl_matrix_vector_mult(pgmR, pgv, pgv);
	    pfN[ul*3  ] = gsl_vector_get(pgv, 0); pfN[ul*3+1] = gsl_vector_get(pgv, 1); pfN[ul*3+2] = gsl_vector_get(pgv, 2);
	}
	gsl_vector_free(pgv);
	gsl_matrix_free(pgmR);
    }
}SROBJ;


class CPSReg
{
public:
    CPSReg();
    virtual ~CPSReg();

    int runRegistration(const int nNSteps, const int nNLMSteps, 
			const float fMESigma, const float fNThresh, 
			const float fTDSearch, const float fCGSearch,
			const bool bBCheck, const bool bIEval, const int nMultiNP,
			const bool bPairwise, bool* pbModels,
			bool** pbScnList);
    void addObject(C3DFile* pc3dObj);
    void updateMatrices(std::vector<GLfloat*>* pvmats);
    void uploadMatrices(std::vector<GLfloat*>* pvmats);
    void updateCoords(){
	std::vector<SROBJ*>::iterator itr;
	for(itr = m_vsrObjs.begin(); itr != m_vsrObjs.end(); itr++){
	    (*itr)->update();
	}
    }
    SROBJ* getObj(const int i){
	return m_vsrObjs[i];
    }
    void clearPMs(){
	std::vector<SROBJ*>::iterator itr;
	for(itr = m_vsrObjs.begin(); itr != m_vsrObjs.end(); itr++){
	    gsl_matrix_set_identity((*itr)->pM);
	}
    }
    void cleanObjects(){
	std::vector<SROBJ*>::iterator itr;
	for(itr = m_vsrObjs.begin(); itr != m_vsrObjs.end(); itr++){
	    delete [] (*itr)->pfC;
	    delete [] (*itr)->pfN;
	    delete [] (*itr)->pfI;
	    delete [] (*itr)->pulF;
	    gsl_matrix_free((*itr)->pM);
	    gsl_matrix_free((*itr)->pRecM);
	}
	m_vsrObjs.clear();
    }
    int dump3DFile(const int nID, C3DFile* pc3df);

private:

    std::vector<SROBJ*> m_vsrObjs;
    CKDTree* m_pkdt;
    
    int createKDTrees();
    int deleteKDTrees();
    int samplePoints(SROBJ* srobj, float fPrcnt);
    int randomSampling(SROBJ* srobj, float fPrcnt);
    
};

typedef struct _argsSKDT{
    int nID;
    pthread_mutex_t* pmutex;
    CKDTree* pkdt;
    SROBJ* pObj;
}argsSKDT;

    
    
#endif // _PSREG_H_
