/**************************************************************************
 *                                                                        *
 *                       This file is part of PSReg                       *
 *                                                                        *
 *          Copyright (C) 2002 Ko Nishino. All Rights Reserved.           *
 *                                                                        *
 *  Permission to use this software for educational and research purposes *
 * is hereby granted without fee. All other rights, including the rights  *
 * to distribute, modify, or sell the software, are explicitly reserved   *
 * by the author. Please contact the author if you wish to make any       *
 * modifications or additions to the software.                            *
 *                                                                        *
 *  This software is provided "as is" and without warranty of any kind,   *
 * express, implied or otherwise, including without limitation, any       *
 * warranty of merchantability or fitness for a particular purpose.       *
 *                                                                        *
 **************************************************************************/
/*  
 *  RegKernel.cpp
 *          Ko Nishino 
 *               kon@cvl.iis.u-tokyo.ac.jp
 *               http://www.cvl.iis.u-tokyo.ac.jp/~kon/
 */

#include "KDefs.h"
#include "Pose.h"
#include "RegKernel.h"
#include "Utils.h"
#include <pthread.h>
#include <math.h>
#include <iostream>
#include <fstream>



void* thrdTotalDiff(void* pvoid);
void* thrdComputeGradient(void* pvoid);

CRegKernel::CRegKernel()
    : m_nMID(0), m_pbScnList(NULL), m_nNSteps(0), m_nNLMSteps(0), m_fMESigma(.0), m_fMESigma2(.0), m_fNThresh(.0),
      m_fTDSearch(.0), m_fCGSearch(.0), m_bBCheck(false), m_bIEval(false), m_nMultiNP(0)
{
    std::cout << "Default constructer should not be called" << std::endl;
}

CRegKernel::CRegKernel(CKDTree* pkdt, const int nModel, const bool* pbScnList, 
		       const int nNSteps, const int nNLMSteps, 
		       const float fMESigma, const float fNThresh, 
		       const float fTDSearch, const float fCGSearch, 
		       const bool bBCheck, 
		       const bool bIEval, const int nMultiNP)
    : m_nMID(nModel), m_nNSteps(nNSteps), m_pbScnList(pbScnList), 
      m_nNLMSteps(nNLMSteps), m_fMESigma(fMESigma), m_fMESigma2(fMESigma*fMESigma), m_fNThresh(fNThresh),
      m_fTDSearch(fTDSearch == .0 ? LARGE_INT : fTDSearch), 
      m_fCGSearch(fCGSearch == .0 ? LARGE_INT : fCGSearch),
      m_bBCheck(bBCheck), m_bIEval(bIEval), m_nMultiNP(nMultiNP)
{

    m_pkdt = pkdt;
    m_fMScale = computeModelScale();

    m_fMaxTrans = 2.5*m_fMScale;
    m_fMaxRot = 3.0;

}
    

CRegKernel::~CRegKernel()
{

}

float CRegKernel::run(CPose& cpP)
{
    cpP.setUnit();
    double dtd(0.0), dlasttd(0.0);
    double dgg(0.0), dlastgg(0.0);
    double dkai(0.0);
    double dscale(0.0);
    int nLMSteps = m_nNLMSteps;

    dscale = 1.0+GOLDENRATIO;
    CPose cpG, cpH, cpHS;

    // First run of TotalDiff
    int nDummy(1); 
    dtd = TotalDiff(cpP, nDummy); 

    for(int nstep = 0; nstep < m_nNSteps; nstep++){
	std::cout << std::endl << m_nMID << ": " << nstep+1 << "/" << m_nNSteps << std::endl;
	if(dtd == 0.0){
	    std::cout << "ftd = 0.0" << std::endl;
	    break;
	}

	dlasttd = dtd;
	computeGradient(cpP, cpG);

	normalizeGradient(cpG);

	dgg = cpG.dot(0, 5);
	dkai = (nstep == 0 ? 0.0 : dgg/dlastgg);
	cpH = cpG + cpH*dkai;

	cpHS = cpH;
	scaleGradient(cpHS);

	dtd = LineMinimization(!nstep, nLMSteps, cpP, cpHS, dkai, dlasttd);
	cpP = cpP + cpHS*dkai;

	if(nstep != 0) dscale *= dkai;
	dlastgg = dgg;
	std::cout << std::endl;
	std::cout << "delta: " << dgg << std::endl;
	std::cout << "Pose: ";
	cpP.print();
	std::cout << std::endl;

	if(cpP.checkZero()){
	    std::cout << "terminating..." << std::endl;
	    break;
	}
    }

    cpP.scale(0.5);
    cpP.setUnit();

    return dtd;
}

double CRegKernel::TotalDiff(const CPose& cpP, int& nStep) const
{
    double dtd(0.0);
    nStep--;
    int nNCorr(0);

#ifdef NONTHREADS

    CNTNT* pMdl = m_pkdt->GetContent(m_nMID);
    CRPoint crpM;
    int nstop, nresult;
    int nAborted(0);
    for(unsigned long ulm = 0; ulm < pMdl->size; ulm++){ // for each point of the model 
	getRPfromCntnt(crpM, pMdl, ulm);
	crpM.transform(cpP);
	for(int nsid = 0; nsid < m_pkdt->GetSize(); nsid++){ // for each scene
	    if(nsid == m_nMID) continue; // skip model
	    if(!m_pbScnList[nsid]) continue; // skip those back side range images
	    float fdist; // distance to the nearest point
	    int nsp; // index of the nearest point
	    CNTNT* pScn = m_pkdt->GetContent(nsid);
	    nstop = (int)fabs((double)pScn->size*log((double)pScn->size)*m_fTDSearch);//LARGE_INT;
	    getNearestPoint(crpM, nsid, &fdist, &nsp, nstop, nresult);

	    if(fdist == MAXFLOAT) continue; // couldn't get a nearest neighbor
	    if(nresult == 0) nAborted++;

	    if(m_bBCheck)
		if(pScn->bound[nsp]) continue;
	    if(!NormalCheck(crpM, pScn, nsp, m_fNThresh)) continue;
	    
	    nNCorr++;
	    
	    dtd += MEstimator(fdist);
	}
    }

#ifdef DTLDEBUG
    std::cout << (nAborted*100.0)/(pMdl->size*(m_pkdt->GetSize()-1)) << "% aborted" << std::endl;
#endif // DTLDEBUG

#endif // NONTHREADS 

#ifdef THREADS
    int nNScns(0);
    for(int i = 0; i < m_pkdt->GetSize(); i++){
	if(m_pbScnList[i]) nNScns++;
    }
    argsTD** pArgs = new (argsTD (*[nNScns]));
    int* pnNCorrs = new int[nNScns];
    double* pdtds = new double[nNScns];
    pthread_t* pthrds = new pthread_t[nNScns];
    pthread_mutex_t mutex;
    pthread_mutex_init(&mutex, NULL);
    int ncnt(0);
    for(int nsid = 0; nsid < m_pkdt->GetSize(); nsid++){
	if(nsid == m_nMID) continue; // skip model
	if(!m_pbScnList[nsid]) continue;
	pArgs[ncnt] = new argsTD;
	pArgs[ncnt]->pmutex = &mutex;
	pArgs[ncnt]->prk = const_cast<CRegKernel*>(this);
	pArgs[ncnt]->cpP = cpP;
	pArgs[ncnt]->pkdt = m_pkdt;
	pArgs[ncnt]->nMID = m_nMID;
	pArgs[ncnt]->nSID = nsid;
	pArgs[ncnt]->bBCheck = m_bBCheck;
	pArgs[ncnt]->fNThresh = m_fNThresh;
	pArgs[ncnt]->fTDSearch = m_fTDSearch;
	pArgs[ncnt]->pnNCorr = &(pnNCorrs[ncnt]);
	pArgs[ncnt]->pdtd = &(pdtds[ncnt]);

	pthread_create(&(pthrds[ncnt]), NULL, thrdTotalDiff, pArgs[ncnt]);
	ncnt++;
    }
    
    for(int i = 0; i < nNScns; i++){
	pthread_join(pthrds[i], NULL);
    }

    for(int i = 0; i < nNScns; i++){
	nNCorr += (*(pArgs[i]->pnNCorr));
	dtd += (*(pArgs[i]->pdtd));
    }

    for(int i = 0; i < nNScns; i++){
	delete pArgs[i];
    }
    delete [] pArgs;
    delete [] pnNCorrs;
    delete [] pdtds;
    delete [] pthrds;

#endif // THREADS    

    if(nNCorr == 0){
	std::cout << "No correspondence" << std::endl;
	return 0.0;
    }else{
	return (dtd/nNCorr);
    }

}

void CRegKernel::getNearestPoint(const CRPoint& crp, const int nsid, float* fdist, int* nsp, int nstop, int& nresult) const
{
    static float pftarget[3];
    CNTNT* pScn = m_pkdt->GetContent(nsid);
    for(int i = 0; i < 3; i++) pftarget[i] = gsl_vector_get(crp.getCoord(), i);
    
    if(!m_bIEval){ // don't evaluate intensity
	m_pkdt->KNNSearch(pScn->data, pftarget, fdist, nsp, pScn->dim, 1, pScn->size, nsid, false, nstop, nresult);
    }else{ // pick some nearest neighbors and eval the intensity
	float *pfdist;
	pfdist = new float[m_nMultiNP];
	int *pnsp;
	pnsp = new int[m_nMultiNP];
	m_pkdt->KNNSearch(pScn->data, pftarget, pfdist, pnsp, pScn->dim, m_nMultiNP, pScn->size, nsid, false, nstop, nresult);
	float fmin, ftmp;
	int nmin;
	fmin = fabs(crp.getInten()-pScn->inten[pnsp[0]]);
	nmin = 0; 
	for(int i = 1; i < m_nMultiNP; i++){
	    if(pnsp[i] == -1) continue;
	    ftmp = fabs(crp.getInten()-pScn->inten[pnsp[i]]);
	    if(ftmp < fmin){
		nmin = i;
		fmin = ftmp;
	    }
	}
	*fdist = pfdist[nmin];
	*nsp = pnsp[nmin];

	delete [] pfdist;
	delete [] pnsp;
    }
}
	
void CRegKernel::computeGradient(const CPose& cpP, CPose& cpG) const
{
    cpG.initialize();
    int nNCorr(0);
    float fweight(0.0);

#ifdef NONTHREADS
    CNTNT* pMdl = m_pkdt->GetContent(m_nMID);
    CRPoint crpM, crpS;
    gsl_vector* vMR = gsl_vector_alloc(3); // rotation only current position = XM
    gsl_vector* vTY = gsl_vector_alloc(3); // t-y_i = XP
    gsl_vector* vGf = gsl_vector_alloc(3);
    gsl_vector* vGb = gsl_vector_alloc(3);
    gsl_vector* vG = gsl_vector_alloc(7);
    int nstop, nresult;
    int nAborted(0);
    
    for(unsigned long ulm = 0; ulm < pMdl->size; ulm++){ // for each point of
							 // the model 
	getRPfromCntnt(crpM, pMdl, ulm);
	crpM.transform(cpP);
	for(int nsid = 0; nsid < m_pkdt->GetSize(); nsid++){ // for each scene
	    if(nsid == m_nMID) continue; // skip model
	    if(!m_pbScnList[nsid]) continue; // skip those back side range images
	    float fdist; // distance to the nearest point
	    int nsp; // index of the nearest point
	    CNTNT* pScn = m_pkdt->GetContent(nsid);
	    nstop = (int)fabs((double)pScn->size*log((double)pScn->size)*m_fCGSearch);

	    getNearestPoint(crpM, nsid, &fdist, &nsp, nstop, nresult);
	    if(fdist == MAXFLOAT) continue; // couldn't get a nearest neighbor

	    if(nresult) nAborted++;
	    if(m_bBCheck)
		if(pScn->bound[nsp]) continue;
	    if(!NormalCheck(crpM, pScn, nsp, m_fNThresh)) continue;
	    
	    nNCorr++;
	    

	    for(int i = 0; i < 3; i++) 
		gsl_vector_set(vMR, i, gsl_vector_get(crpM.getCoord(), i)-gsl_vector_get(cpP.getPose(), i));
	    getRPfromCntnt(crpS, pScn, nsp);

	    fweight = DMEstimator(fdist);

	    for(int i = 0; i < 3; i++)
		gsl_vector_set(vTY, i, gsl_vector_get(cpP.getPose(), i)-gsl_vector_get(crpS.getCoord(), i));

	    gsl_vector_memcpy(vGf, vMR);
	    gsl_vector_add(vGf, vTY);
	    gsl_vector_scale(vGf, 2.0*fweight);
	    
	    Utils::gsl_vector3_cross(vMR, vTY, vGb);
	    gsl_vector_scale(vGb, 4.0*fweight);
	    
	    gsl_vector_set_zero(vG);
	    for(int i = 0; i < 3; i++)
		gsl_vector_set(vG, i, gsl_vector_get(vGf, i));
	    for(int i = 3; i < 6; i++)
		gsl_vector_set(vG, i, gsl_vector_get(vGb, i-3));

	    cpG.addVector(vG);

	}
    }
    gsl_vector_free(vMR);
    gsl_vector_free(vTY);
    gsl_vector_free(vGf);
    gsl_vector_free(vGb);
    gsl_vector_free(vG);

#ifdef DTLDEBUG
    std::cout << (nAborted*100.0)/(pMdl->size*(m_pkdt->GetSize()-1)) << "% aborted" << std::endl;
#endif // DTLDEBUG

#endif // NONTHREADS

#ifdef THREADS
    int nNScns(0);
    for(int i = 0; i < m_pkdt->GetSize(); i++){
	if(m_pbScnList[i]) nNScns++;
    }
    argsCG** pArgs = new (argsCG (*[nNScns]));
    int* pnNCorrs = new int[nNScns];
    double* pdtds = new double[nNScns];
    CPose* pcpGs = new CPose[nNScns];
    pthread_t* pthrds = new pthread_t[nNScns];
    pthread_mutex_t mutex;
    pthread_mutex_init(&mutex, NULL);
    int ncnt(0);
    for(int nsid = 0; nsid < m_pkdt->GetSize(); nsid++){
	if(nsid == m_nMID) continue; // skip model
	if(!m_pbScnList[nsid]) continue;
	pArgs[ncnt] = new argsCG;
	pArgs[ncnt]->pmutex = &mutex;
	pArgs[ncnt]->prk = const_cast<CRegKernel*>(this);
	pArgs[ncnt]->cpP = cpP;
	pArgs[ncnt]->pkdt = m_pkdt;
	pArgs[ncnt]->nMID = m_nMID;
	pArgs[ncnt]->nSID = nsid;
	pArgs[ncnt]->bBCheck = m_bBCheck;
	pArgs[ncnt]->fNThresh = m_fNThresh;
	pArgs[ncnt]->fCGSearch = m_fCGSearch;
	pArgs[ncnt]->pnNCorr = &(pnNCorrs[ncnt]);
	pArgs[ncnt]->pcpG = &(pcpGs[ncnt]);

	pthread_create(&(pthrds[ncnt]), NULL, thrdComputeGradient, pArgs[ncnt]);
	ncnt++;
    }

    for(int i = 0; i < nNScns; i++){
	pthread_join(pthrds[i], NULL);
    }

    for(int i = 0; i < nNScns; i++){
	nNCorr += (*(pArgs[i]->pnNCorr));
	cpG = cpG + (*(pArgs[i]->pcpG));
    }

    for(int i = 0; i < nNScns; i++){
	delete pArgs[i];
    }
    delete [] pArgs;
    delete [] pnNCorrs;
    delete [] pdtds;
    delete [] pthrds;
    delete [] pcpGs;

#endif // THREADS
    
    if(nNCorr == 0){
	cpG.initialize();
    }else{
	cpG.scale(-1.0/nNCorr);
    }
}

float CRegKernel::computeModelScale() const
{
    float fmaxX, fmaxY, fmaxZ, fminX, fminY, fminZ;
    fmaxX = fmaxY = fmaxZ = -1000.0;
    fminX = fminY = fminZ = 1000.0;
 
    CNTNT* pMdl = m_pkdt->GetContent(m_nMID);    
    for(unsigned long id = 0; id < pMdl->size; id++){
	fmaxX = fmaxX > pMdl->data[id*3  ] ? fmaxX : pMdl->data[id*3  ];
	fmaxY = fmaxY > pMdl->data[id*3+1] ? fmaxY : pMdl->data[id*3+1];
	fmaxZ = fmaxZ > pMdl->data[id*3+2] ? fmaxZ : pMdl->data[id*3+2];
	fminX = fminX < pMdl->data[id*3  ] ? fminX : pMdl->data[id*3  ];
	fminY = fminY < pMdl->data[id*3+1] ? fminY : pMdl->data[id*3+1];
	fminZ = fminZ < pMdl->data[id*3+2] ? fminZ : pMdl->data[id*3+2];
    }
    
    float fscaleX, fscaleY, fscaleZ, ftmp;
    fscaleX = fmaxX - fminX;
    fscaleY = fmaxY - fminY;
    fscaleZ = fmaxZ - fminZ;
    ftmp = fscaleY > fscaleZ ? fscaleY : fscaleZ;
    
    return fscaleX > ftmp ? fscaleX : ftmp;
}

void CRegKernel::scaleGradient(CPose& cpHS) const
{
    float fscale, fscale1, fscale2;
    fscale = m_fMScale;
    float fmaxtrans, fmaxrot;
    fmaxtrans = m_fMaxTrans;
    fmaxrot = sin((double)m_fMaxRot*DEG2RAD*0.5);
    cpHS.scale(0, 2, fscale);

    fscale1 = fmaxtrans/sqrt(cpHS.dot(0, 2));
    fscale2 = fmaxrot/sqrt(cpHS.dot(3, 5));
    fscale = fscale1 < fscale2 ? fscale1 : fscale2;
    
    cpHS = cpHS*fscale;
}
    

double CRegKernel::LineMinimization(int firsttime, int nstep,
				    CPose& cpP, CPose& cpG,
				    double& dkai, double dlasttd) const
{
    double dtdA, dtdB, dtdC, dtdX, dA, dB, dC, dX;
    CPose cpNP;
    
    dtdA = dlasttd;
    if(firsttime){
	MustBracketMin(cpP, cpG, dA, dB, dC, dtdA, dtdB, dtdC);
    }else{
	BracketMin(cpP, cpG, dA, dB, dC, dtdA, dtdB, dtdC, nstep);
    }

    if(QuadraticMin(cpP, cpG, dA, dB, dC, dtdA, dtdB, dtdC, dX, dtdX, nstep)){
	dkai = dX;
	return dtdX;
    }

    while(nstep > 0){
	if(fabs(dC-dB) > fabs(dB-dA)){
	    dX = (dC-dB)*GOLDENRATIO2+dB;
	    cpNP = cpP + cpG*dX;
	    dtdX = TotalDiff(cpNP, nstep);
	    if(dtdX < dtdB){
		dA = dB; dtdA = dtdB;
		dB = dX; dtdB = dtdX;
	    }else{
		dC = dX; dtdC = dtdX;
	    }
	}else{
	    dX = (dA-dB)*GOLDENRATIO2+dB;
	    cpNP = cpP + cpG*dX;
	    dtdX = TotalDiff(cpNP, nstep);
	    if(dtdX < dtdB){
		dC = dB; dtdC = dtdB;
		dB = dX; dtdB = dtdX;
	    }else{
		dA = dX; dtdA = dtdX;
	    }
	}
    }
    
    dkai = (dtdB < dtdC) ? ((dtdB < dtdA) ? dB : dA) : ((dtdA < dtdC) ? dA : dC);

    return ((dtdB < dtdC) ? ((dtdB < dtdA) ? dtdB : dtdA) : ((dtdA < dtdC) ? dtdA : dtdC));
}


void CRegKernel::BracketMin(const CPose& cpP, const CPose& cpG,
			    double& dA, double& dB, double& dC,
			    double& dtdA, double& dtdB, double& dtdC, int& nstep) const
{
    CPose cpNP;
    
    dA = 0.0; dB = GOLDENRATIO2; dC = 1.0;
    cpNP = cpP + cpG*dB;
    dtdB = TotalDiff(cpNP, nstep);
    
    if(dtdA < dtdB){
	while((dtdA < dtdB) && (nstep > 0)){
	    dtdC = dtdB; dC = dB;
	    dB = (dB-dA)*GOLDENRATIO2+dA;
	    cpNP = cpP + cpG*dB;
	    dtdB = TotalDiff(cpNP, nstep);
	}
    }else{
	cpNP = cpP + cpG*dC;
	dtdC = TotalDiff(cpNP, nstep);
    }

    if(dtdC < dtdB){
	while((dtdC < dtdB) && (nstep > 0)){
	    dtdA = dtdB; dA = dB; 
	    dtdB = dtdC; dB = dC; 
	    dC = (dB-dA)*GOLDENRATIO2/GOLDENRATIO+dB;
	    cpNP = cpP + cpG*dC;
	    dtdC = TotalDiff(cpNP, nstep);
	}
    }
}

void CRegKernel::MustBracketMin(const CPose& cpP, const CPose& cpG,
				double& dA, double& dB, double& dC, 
				double& dtdA, double& dtdB, double& dtdC) const
{
    int nstep = MAXBRACKET_STEPS;
    dA = 0.0; dB = GOLDENRATIO2; dC = 1.0;
    CPose cpNP;
    cpNP = cpP + cpG*dB;
    dtdB = TotalDiff(cpNP, nstep);
    
    if(dtdA < dtdB){
	while((dtdA < dtdB) && (nstep > 0)){
	    dtdC = dtdB; dC = dB;
	    dB = (dB-dA)*GOLDENRATIO2+dA;
	    cpNP = cpP + cpG*dB;
	    dtdB = TotalDiff(cpNP, nstep);
	}
    }else{
	cpNP = cpP + cpG*dC;
	dtdC = TotalDiff(cpNP, nstep);
    }
    
    if(dtdC < dtdB){
	while((dtdC < dtdB) && (nstep > 0)){
	    dtdA = dtdB; dA = dB; 
	    dtdB = dtdC; dB = dC;
	    dC += GOLDENRATIO;
	    cpNP = cpP + cpG*dC;
	    dtdC = TotalDiff(cpNP, nstep);
	}
    }
}

int CRegKernel::QuadraticMin(const CPose& cpP, const CPose& cpG, 
			     double dA, double dB, double dC,
			     double dtdA, double dtdB, double dtdC,
			     double& dX, double& dtdX, int& nstep) const
{
    CPose cpNP;

    double dmother;
    dmother = -((dB-dC)*(-dtdA+dtdB))+(-dA+dB)*(dtdB-dtdC);
    if(fabs(dmother) < 0.0000000000000001){
	dX = dB;
    }else{
	dX = dB-0.5*(-((dB-dC)*(dB-dC)*(-dtdA+dtdB))+(-dA+dB)*(-dA+dB)*(dtdB-dtdC))/dmother;
    }
    cpNP = cpP + cpG*dX;
    dtdX = TotalDiff(cpNP, nstep);
    
    if(dX > 4.0*dC){
    }

    if((dtdX < dtdA) && (dtdX < dtdB) && (dtdX < dtdC)){
	return 1;
    }else{
	return 0;
    }
}

// Threads 


void* thrdTotalDiff(void* pvoid)
{
    argsTD* pargs = (argsTD*)pvoid;
    // initialize output values
    *(pargs->pnNCorr) = 0;
    *(pargs->pdtd) = 0.0;
#ifdef DTLDEBUG
    std::cout << "thrdTotalDiff: " << pargs->nSID << " hooked";
#endif // DTLDEBUG
    CNTNT* pMdl = pargs->pkdt->GetContent(pargs->nMID);
    CNTNT* pScn = pargs->pkdt->GetContent(pargs->nSID);
    CRPoint crpM;
    float fdist;
    int nsp;
    int nstop = (int)fabs((double)pScn->size*log((double)pScn->size)*pargs->fTDSearch);
#ifdef DTLDEBUG
    std::cout << " " << pMdl->size << std::endl;
    std::cout << "searching range: " << nstop << std::endl;
#endif // DTLDEBUG
    int nresult(-1);
    int nAborted(0);
    for(unsigned long ulm = 0; ulm < pMdl->size; ulm++){ // for each point of the model
	pargs->prk->getRPfromCntnt(crpM, pMdl, ulm);
	crpM.transform(pargs->cpP);

	pthread_mutex_lock(pargs->pmutex);
	pargs->prk->getNearestPoint(crpM, pargs->nSID, &fdist, &nsp, nstop, nresult);
	pthread_mutex_unlock(pargs->pmutex);

	if(fdist == MAXFLOAT) continue; // couldn't get a nearest neighbor
	if(nresult == 0){
	    // although it aborted, it has a closest neighbor! USE THAT
	    nAborted++;
	}

	if(pargs->bBCheck)
	    if(pScn->bound[nsp]) continue;
	if(!pargs->prk->NormalCheck(crpM, pScn, nsp, pargs->fNThresh)) continue;

	(*(pargs->pnNCorr))++;
	*(pargs->pdtd) += pargs->prk->MEstimator(fdist);
    }

#ifdef DTLDEBUG
    std::cout << "thrdTotalDiff: " << pargs->nSID << "-> "
	      << *(pargs->pnNCorr) << "\t" << *(pargs->pdtd) << " "
	      << nAborted*100.0/pMdl->size << "% aborted " 
	      << *(pargs->pnNCorr)*100.0/pMdl->size << "% used" << std::endl;
#endif // DTLDEBUG
#ifndef DTLDEBUG
    std::cout << "*" << std::flush;
#endif
	
    return NULL;
}

void* thrdComputeGradient(void* pvoid)
{
    argsCG* pargs = (argsCG*)pvoid;
    // initialize output values
    *(pargs->pnNCorr) = 0;
    pargs->pcpG->initialize();
#ifdef DTLDEBUG
    std::cout << "thrdComputeGradient: " << pargs->nSID << " hooked";
#endif // DTLDEBUG
    CNTNT* pMdl = pargs->pkdt->GetContent(pargs->nMID);
    CNTNT* pScn = pargs->pkdt->GetContent(pargs->nSID);
    CRPoint crpM, crpS;
    float fdist, fweight;
    int nsp;
    gsl_vector* vMR = gsl_vector_alloc(3); // rotation only current position = XM
    gsl_vector* vTY = gsl_vector_alloc(3); // t-y_i = XP
    gsl_vector* vGf = gsl_vector_alloc(3);
    gsl_vector* vGb = gsl_vector_alloc(3);
    gsl_vector* vG = gsl_vector_alloc(7);
    int nstop = (int)fabs((double)pScn->size*log((double)pScn->size)*pargs->fCGSearch);

#ifdef DTLDEBUG
    std::cout << " " << pMdl->size << std::endl;
    std::cout << "searching range: " << nstop << std::endl;
#endif // DTLDEBUG
    int nresult;
    int nAborted(0);
    for(unsigned long ulm = 0; ulm < pMdl->size; ulm++){ // for each point of the model
	pargs->prk->getRPfromCntnt(crpM, pMdl, ulm);
	crpM.transform(pargs->cpP);

	pthread_mutex_lock(pargs->pmutex);
	pargs->prk->getNearestPoint(crpM, pargs->nSID, &fdist, &nsp, nstop, nresult);
	pthread_mutex_unlock(pargs->pmutex);

	if(fdist == MAXFLOAT) continue; // couldn't get a nearest neighbor
	if(nresult == 0){
	    nAborted++;
	}

	if(pargs->bBCheck)
	    if(pScn->bound[nsp]) continue;
	if(!pargs->prk->NormalCheck(crpM, pScn, nsp, pargs->fNThresh)) continue;

	(*(pargs->pnNCorr))++;

	for(int i = 0; i < 3; i++) 
	    gsl_vector_set(vMR, i, gsl_vector_get(crpM.getCoord(), i)-gsl_vector_get(pargs->cpP.getPose(), i));

	pargs->prk->getRPfromCntnt(crpS, pScn, nsp);

	fweight = pargs->prk->DMEstimator(fdist);
	
	for(int i = 0; i < 3; i++)
	    gsl_vector_set(vTY, i, gsl_vector_get(pargs->cpP.getPose(), i)-gsl_vector_get(crpS.getCoord(), i));
	
	gsl_vector_memcpy(vGf, vMR);
	gsl_vector_add(vGf, vTY);
	gsl_vector_scale(vGf, 2.0*fweight);
	
	Utils::gsl_vector3_cross(vMR, vTY, vGb);
	gsl_vector_scale(vGb, 4.0*fweight);
	
	gsl_vector_set_zero(vG);
	for(int i = 0; i < 3; i++)
	    gsl_vector_set(vG, i, gsl_vector_get(vGf, i));
	for(int i = 3; i < 6; i++)
	    gsl_vector_set(vG, i, gsl_vector_get(vGb, i-3));
	gsl_vector_set(vG, 6, 0.0);
	
	pargs->pcpG->addVector(vG);
	
    }

#ifdef DTLDEBUG
    std::cout << "thrdComputeGradient: " << pargs->nSID << "-> "
	      << *(pargs->pnNCorr) << " "
	      << nAborted*100.0/pMdl->size << "% aborted " 
	      << *(pargs->pnNCorr)*100.0/pMdl->size << "% used" << std::endl;
    std::cout << "gradient ";
    pargs->pcpG->print();
#endif // DTLDEBUG
#ifndef DTLDEBUG
    std::cout << "+" << std::flush;
#endif
	
    return NULL;
}

