/**************************************************************************
 *                                                                        *
 *                       This file is part of PSReg                       *
 *                                                                        *
 *          Copyright (C) 2002 Ko Nishino. All Rights Reserved.           *
 *                                                                        *
 *  Permission to use this software for educational and research purposes *
 * is hereby granted without fee. All other rights, including the rights  *
 * to distribute, modify, or sell the software, are explicitly reserved   *
 * by the author. Please contact the author if you wish to make any       *
 * modifications or additions to the software.                            *
 *                                                                        *
 *  This software is provided "as is" and without warranty of any kind,   *
 * express, implied or otherwise, including without limitation, any       *
 * warranty of merchantability or fitness for a particular purpose.       *
 *                                                                        *
 **************************************************************************/
/*  
 *  RegKernel.h
 *          Ko Nishino 
 *               kon@cvl.iis.u-tokyo.ac.jp
 *               http://www.cvl.iis.u-tokyo.ac.jp/~kon/
 */

#ifndef _REGKERNEL_H_
#define _REGKERNEL_H_

#include <pthread.h>
#include <math.h>
#include "KDTree.h"
#include "RPoint.h"


#define GOLDENRATIO 0.6180339887498948482
#define GOLDENRATIO2 0.3819660112501051518
#define MAXBRACKET_STEPS 10

    
class CRegKernel
{
public:
    CRegKernel();
    CRegKernel(CKDTree* pkdt, const int nModel, const bool* pbScnList, 
	       const int nNSteps = 3, const int nNLMSteps = 5, 
	       const float fMESigma = .1, const float fNThresh = -1.0, 
	       const float fTDSearch = .01, const float fCGSearch = .01,
	       const bool bBCheck = false, 
	       const bool bIEval = false, const int nMultiNP = 10);
    virtual ~CRegKernel();

    float run(CPose& cp);

private:
    CKDTree* m_pkdt;
    const int m_nMID;
    const bool* m_pbScnList;
    const int m_nNSteps; // number of iterations
    const int m_nNLMSteps; // number of iterations for line minimization
    const float m_fMESigma; // M-estimator's sigma
    const float m_fMESigma2;
    const float m_fNThresh; // threshold on normal angles
    const bool m_bIEval; // whether to eval intensity distance
    const int m_nMultiNP; // how many nearest neighbor candidates to use when m_bIEval==true
    const bool m_bBCheck; // whether to check boundary vertices
    const float m_fTDSearch, m_fCGSearch;
    float m_fMScale, m_fMaxTrans, m_fMaxRot;

    float computeModelScale() const;
    double TotalDiff(const CPose& cpP, int& nStep) const;

    double LineMinimization(int firsttime, int nstep, CPose& cpP, CPose& cpG, double& dkai, double dlasttd) const;
    void BracketMin(const CPose& cpP, const CPose& cpG, double& dA, double& dB, double& dC, double& dtdA, double& dtdB, double& dtdC, int& nstep) const;
    void MustBracketMin(const CPose& cpP, const CPose& cpG, double& dA, double& dB, double& dC, double& dtdA, double& dtdB, double& dtdC) const;
    int QuadraticMin(const CPose& cpP, const CPose& cpG, double dA, double dB, double dC, double dtdA, double dtdB, double dtdC, double& dX, double& dtdX, int& nstep) const;
    void computeGradient(const CPose& cpP, CPose& cpG) const;
    void scaleGradient(CPose& cpHS) const;
    inline void normalizeGradient(CPose& cpG) const;

public: // to enable access from thrdTotalDiff
    void getNearestPoint(const CRPoint& crp, const int nsid, float* fdist, int* nsp, int nstop, int& nresult) const;
    inline void getRPfromCntnt(CRPoint& crp, const CNTNT* pcntnt, const unsigned long ulid) const;
    inline bool NormalCheck(const CRPoint& crp, const CNTNT* pcntnt, const int nid, const float fthresh) const;

    float MEstimator(float fdist) const { return Lorentzian(fdist); }
    inline float Lorentzian(float fdist) const;

    float DMEstimator(float fdist) const { return DLorentzian(fdist); }
    inline float DLorentzian(float fdist) const;
};

typedef struct _argsTD{
    /// input args
    pthread_mutex_t* pmutex;
    CRegKernel* prk; // to use getRPfromCntnt, getNearestPoint, NormalCheck, MEstimator
    CPose cpP;
    CKDTree* pkdt; // KDTree to search in
    int nMID; // Model ID
    int nSID; // Scene ID
    bool bBCheck;
    float fNThresh, fTDSearch; 
    /// output 
    int* pnNCorr;
    double* pdtd;
}argsTD;

typedef struct _argsCG{
    /// input args
    pthread_mutex_t* pmutex;
    CRegKernel* prk; // to use getRPfromCntnt, getNearestPoint, NormalCheck, MEstimator
    CPose cpP;
    CKDTree* pkdt; // KDTree to search in
    int nMID; // Model ID
    int nSID; // Scene ID
    bool bBCheck;
    float fNThresh, fCGSearch; 
    /// output 
    int* pnNCorr;
    CPose* pcpG;
}argsCG;

#include "RegKernel.inl"

#endif // _REGKERNEL_H_
    


