/**************************************************************************
 *                                                                        *
 *                       This file is part of PSReg                       *
 *                                                                        *
 *          Copyright (C) 2002 Ko Nishino. All Rights Reserved.           *
 *                                                                        *
 *  Permission to use this software for educational and research purposes *
 * is hereby granted without fee. All other rights, including the rights  *
 * to distribute, modify, or sell the software, are explicitly reserved   *
 * by the author. Please contact the author if you wish to make any       *
 * modifications or additions to the software.                            *
 *                                                                        *
 *  This software is provided "as is" and without warranty of any kind,   *
 * express, implied or otherwise, including without limitation, any       *
 * warranty of merchantability or fitness for a particular purpose.       *
 *                                                                        *
 **************************************************************************/
/*  
 *  Utils.inl
 *          Ko Nishino 
 *               kon@cvl.iis.u-tokyo.ac.jp
 *               http://www.cvl.iis.u-tokyo.ac.jp/~kon/
 */
// inline function of Utils

inline void Utils::gsl_vector_normalize(gsl_vector* v)
{
    double dMag(0.0);
    for(int i = 0; i < v->size; i++){
	dMag += (gsl_vector_get(v, i)*gsl_vector_get(v, i));
    }

    if(dMag == 0.0){
	gsl_vector_set_zero(v);
    }else{
	assert(dMag > 0.0);
	dMag = sqrt(dMag);
	gsl_vector_scale(v, 1.0/dMag);
    }
}

inline void Utils::gsl_vector3_cross(const gsl_vector* vl, const gsl_vector* vr, gsl_vector* v)
{
    assert(vl->size == 3 && vr->size == 3 && v->size == 3);

    gsl_vector_set(v, 0, gsl_vector_get(vl, 1)*gsl_vector_get(vr, 2)-gsl_vector_get(vl, 2)*gsl_vector_get(vr, 1));
    gsl_vector_set(v, 1, gsl_vector_get(vl, 2)*gsl_vector_get(vr, 0)-gsl_vector_get(vl, 0)*gsl_vector_get(vr, 2));
    gsl_vector_set(v, 2, gsl_vector_get(vl, 0)*gsl_vector_get(vr, 1)-gsl_vector_get(vl, 1)*gsl_vector_get(vr, 0));
}

inline double Utils::gsl_vector_dot(const int nF, const int nT, const gsl_vector* v)
{
    double dsum(0.0);
    for(int i = nF; i <= nT; i++){
	dsum += (gsl_vector_get(v, i)*gsl_vector_get(v, i));
    }
    
    return dsum;
}

inline void Utils::gsl_matrix_diff(gsl_matrix* oldm, gsl_matrix* newm, gsl_matrix* m)
{
    
}

inline void Utils::gsl_matrix_vector_mult(gsl_matrix* m, gsl_vector* v, gsl_vector* vr)
{
    assert(m->size2 == v->size);
    assert(m->size1 == vr->size);

    gsl_vector* vtmp = gsl_vector_alloc(vr->size);
    gsl_vector_set_zero(vtmp);

    double dtmp(0.0);
    for(int i = 0; i < vr->size; i++){
	dtmp = 0.0;
	for(int j = 0; j < v->size; j++){
	    dtmp += gsl_matrix_get(m, i, j)*gsl_vector_get(v, j);
	}
	gsl_vector_set(vtmp, i, dtmp);
    }
    
    gsl_vector_memcpy(vr, vtmp);
    gsl_vector_free(vtmp);
}

inline void Utils::gsl_matrix4x4_mult(gsl_matrix* ml, gsl_matrix* mr, gsl_matrix* m)
{
    assert(ml->size1 == 4 && ml->size2 == 4);
    assert(mr->size1 == 4 && mr->size2 == 4);
    assert(m->size1 == 4 && m->size2 == 4);

    float pf[4][4];
    for(int i = 0; i < 4; i++){
	for(int j = 0; j < 4; j++){
	    pf[i][j] = 0.0;
	}
    }
    for(int i = 0; i < 4; i++){
	for(int j = 0; j < 4; j++){
	    for(int k = 0; k < 4; k++){
		pf[i][j] += gsl_matrix_get(ml, i, k)*gsl_matrix_get(mr, k, j);
	    }
	}
    }
    for(int i = 0; i < 4; i++){
	for(int j = 0; j < 4; j++){
	    gsl_matrix_set(m, i, j, pf[i][j]);
	}
    }
}
	    
inline void Utils::gsl_matrix4x4_from_pfloat(gsl_matrix* m, const float* pglf, const bool btrans)	
{
    assert(m->size1 == 4);
    assert(m->size2 == 4);
    
    for(int i = 0; i < 4; i++){
	for(int j = 0; j < 4; j++){
	    if(btrans){
		gsl_matrix_set(m, i, j, pglf[i+j*4]);
	    }else{
		gsl_matrix_set(m, i, j, pglf[i*4+j]);
	    }
	}
    }
}

inline void Utils::gsl_matrix4x4_to_pfloat(const gsl_matrix* m, float* pglf, const bool btrans)
{
    assert(m->size1 == 4);
    assert(m->size2 == 4);
    
    for(int i = 0; i < 4; i++){
	for(int j = 0; j < 4; j++){
	    if(btrans){
		pglf[i+j*4] = gsl_matrix_get(m, i, j);
	    }else{
		pglf[i*4+j] = gsl_matrix_get(m, i, j);
	    }
	}
    }
}

    

inline void Utils::Intensity2RGB(float* pC, float fInten, float* pBaseC)
{
    float fBR, fBG, fBB;
    fBR = pBaseC[0]*256.0; fBG = pBaseC[1]*256.0; fBB = pBaseC[2]*256.0;
    float fY, fU, fV;
    fY =  0.29900*fBR+0.58700*fBG+0.11400*fBB;
    fU = -0.16874*fBR-0.33126*fBG+0.50000*fBB;
    fV =  0.50000*fBR-0.41869*fBG-0.08131*fBB;

    fY = 256.0*fInten;
    float fR, fG, fB;
    fR = fY           +1.40200*fV;
    fG = fY-0.34414*fU-0.71414*fV;
    fB = fY+1.77200*fU;
    fR = fR/256.0; fG = fG/256.0; fB = fB/256.0;
    fR = fR > 1.0 ? 1.0 : fR; fR = fR < 0.0 ? 0.0 : fR;
    fG = fG > 1.0 ? 1.0 : fG; fG = fG < 0.0 ? 0.0 : fG;
    fB = fB > 1.0 ? 1.0 : fB; fB = fB < 0.0 ? 0.0 : fB;
    pC[0] = fR;
    pC[1] = fG;
    pC[2] = fB;

    pC[3] = 1.0;
}

